/*
   This file is part of the MoCA video library.
   Name: FaceDetectionTest.cpp
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: kopf@informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

/*! \file

 
  
   \author    Stephan Kopf
   
	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de
	
	 (c) Lehrstuhl fuer Praktische Informatik IV, details in COPYRIGHT.txt
	 
*/

#include <iostream>
using namespace std;


#include <FaceDetection.h>
#include <FileList.h>
#include <MpegStatus.h>
#include <DrawDot.h>
#include <LightingCorrection.h>


void usage() 
{
	cerr << "Usage: FaceDetection [options] [JPG|PGM|PNM|PNG|MPG]filename|path" << endl;
	cerr << "Version 0.98" << endl;
	cerr << endl;
	
	cerr << "options: " << endl;
	cerr << "  -recurse                analyses all MPG and AVI files in path" << endl;
	cerr << "  -verbose                report the work to standard output" << endl;
	cerr << "  -size=MEDIUM            search only medium or large faces (faster)" << endl;
	cerr << "  -size=LARGE             search only large faces (much faster)" << endl;
	cerr << "  -size=HUGE              search only huge faces (much faster!)" << endl;
	cerr << "  -skip <int>             skip frames, e.g. -skip 2 analyses every 3rd frame" << endl;
	cerr << "  -one <int>              analyse only one frame, e.g. -one 4 analyses frame 4 (first frame=0)" << endl;
	cerr << "  -rotate                 find rotated faces" << endl;
	cerr << "  -store                  store frames with faces" << endl;
	cerr << endl;
	
	cerr << "Example:" <<  endl;
	cerr << "FaceDetection test.jpg" << endl;
	cerr << "FaceDetection -one test.mpg" << endl;
	cerr << "FaceDetection -recurse /opt/movies/" << endl;
	
	exit(1);
	return;
} // usage


void processVideo (string &fn, FaceDetectionParameter &param, bool storeFrames)
{
	param.inputData = FACE_INPUT_VIDEO;
	param.mpgFilename = fn;
	cout << endl  << "processing file: "<< param.mpgFilename << endl;

	// start face detection
	FaceData data;
	FaceDetection fd(param);
	fd.processVideo(data);
	cout << "number of faces: " << data.faces.size() << endl; 
	//cout << data;
	if (storeFrames) {
		cerr << "store frames" << endl;
		fd.storeFaces (data, fn);
	} // if

	return;
}

void processImage (string &fn, FaceDetectionParameter &param, bool storeFrames)
{
	// read image
	Image<uint8> img;
	ReadImage (fn, img);
	
	// calculate faces
	param.inputData = FACE_INPUT_IMAGE;
	FaceDetection fd(param);
	list<Face> faces;
	fd.processImage(img, faces);
	cerr << faces << endl;
	
	if (storeFrames) {
		
		list<Face>::iterator it;
		for (it=faces.begin(); it!=faces.end(); it++) {
			Face f = (*it);
			
			// define face data
			FaceItem item;
			item.angle = f.angle;
			item.frame=0;
			item.p1[DIM_X] = f.x - f.width / 2;
			item.p1[DIM_Y] = f.x + f.width / 2;
			item.p2[DIM_X] = item.p1[DIM_X] + f.width;
			item.p2[DIM_Y] = item.p1[DIM_Y] + f.width;
			
			// mark faces
			fd.markFace (item, img);
		} // if
		
		// store image
		CreateMocaPath mocaPath (fn, false);
		string fn1 = mocaPath.file.substr(0,mocaPath.file.size()-4); 
		string fn2 = mocaPath.path + "face_data_" + fn1 + ".png";
		WriteImage (img, fn2);
	} // if
	

	return;
}


int main(int argc, char* argv[])
{


    if (argc<2) usage();
	string filename = argv[argc-1];

	bool recurse     = false;
	bool verbose     = false;
	bool mediumFace  = false;
	bool largeFace   = false;
	bool hugeFace    = false;
	bool skip        = false;
	int  skipV       = -1;
	bool one         = false;
	int  oneV        = -1;
	bool rotate      = false;
	bool storeFrames = false;

	string s;
	for (int i=1; i<argc-1; i++) {
		s=argv[i];
		for (uint8 k=0; k<s.size(); k++)
		  s[k]=toupper(s[k]);
		
		if (s=="-RECURSE")          recurse=true;
		else if (s=="-VERBOSE")     verbose=true;
		else if (s=="-SIZE=MEDIUM") mediumFace=true;
		else if (s=="-SIZE=LARGE")  largeFace=true;
		else if (s=="-SIZE=HUGE")   hugeFace=true;
		else if (s=="-ROTATE")      rotate=true;
		else if (s=="-STORE")       storeFrames=true;
		else if (s=="-SKIP") {
			skip=true;
			i++;
			skipV = atoi(argv[i]);
		} else if (s=="-ONE") {
			one=true;
			i++;
			oneV = atoi(argv[i]);
		} else if (i != argc-1) usage();
	} // for
	
	
	FaceDetectionParameter param;
	param.verbose = verbose;
	if (param.verbose) {
		param.outFile=true;
	} // if
	param.useColorInfo = true;
	if (skip)   param.stepWidth = skipV+1;
	if (one)    param.analyseOne=oneV;
	if (rotate) {
		param.rotate=true;
		param.degree = 15;      // 15 degree rotation
		param.rotationStep = 1; // rotate left and right once
	} // if
	if (mediumFace) param.defaultFaceSize=FACE_DETECTION_SIZE_MEDIUM;
	if (largeFace)  param.defaultFaceSize=FACE_DETECTION_SIZE_LARGE;
	if (hugeFace)   param.defaultFaceSize=FACE_DETECTION_SIZE_HUGE;


	bool isImage = false;
	bool isVideo = false;
	string fileExt = filename.substr(filename.size()-4);
	for (uint8 k=0; k<fileExt.size(); k++)
		fileExt[k]=toupper(fileExt[k]);
	if (fileExt == ".JPG") isImage = true;
	if (fileExt == ".PNG") isImage = true;
	if (fileExt == ".PGM") isImage = true;
	if (fileExt == ".PNM") isImage = true;
	if (fileExt == ".MPG") isVideo = true;
	if (fileExt == ".MPEG")isVideo = true;
	if (fileExt == ".AVI") isVideo = true;

	if (!isImage && !isVideo && !recurse) {
		cerr << "unknown image or video: " << filename << endl << endl;
		exit(1);
	}


	// process one image
	if (isImage) {
		processImage(filename, param, storeFrames);
		exit(0);
	} // if
	

	// process one video
	if (isVideo) {
		processVideo(filename, param, storeFrames);
		exit(0);
	} // if

	
	// recurse all videos in directory
	if (recurse) {
		{
			// read all MPG files in directory
			vector<FileInfo> mpgFiles;
			FileList fl(filename, true);    
			fl.process("*.mpg", mpgFiles);
			for (int32 i=0; i<mpgFiles.size(); i++) {
				string fn = mpgFiles[i].name;
				processVideo(fn, param, storeFrames);
			} // for
			cout << endl << mpgFiles.size() << " files found." << endl;
		}
		{
			// read all AVIfiles in directory
			vector<FileInfo> aviFiles;
			FileList fl(filename, true);    
			fl.process("*.avi", aviFiles);
			for (int32 i=0; i<aviFiles.size(); i++) {
				string fn = aviFiles[i].name;
				processVideo(fn, param, storeFrames);
			} // for
			cout << endl << aviFiles.size() << " files found." << endl;
		}
	} // else
	
	return 0;
} // main

