/*
   This file is part of the MoCA video library.
   Name: VideoConverter.cpp
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: kopf@informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

/*! \file

  
	
	  \author    Stephan Kopf
	  
		Lehrstuhl fuer Praktische Informatik IV
		Universitaet Mannheim
		68131 Mannheim
		Germany
		email: kopf@informatik.uni-mannheim.de
		
		  (c) Lehrstuhl fuer Praktische Informatik IV, details in COPYRIGHT.txt
		  
*/


// C++ includes
#include <string>
#include <iostream>
using namespace std;

// MoCA includes
#include<VideoIndex.h>
#include<VideoDecoder.h>
#include<OMpegVideo.h>
#include<VideoEncoder.h>
#include<FlipFilter.h>
#include<RotationFilter.h>


void parseParameters (int argc, char* argv[], VideoEncoderOptions &encoderOpt, int32 &rotate, bool &mirror)
{
	// analyze params
	rotate = 0;
	mirror = false;
	string encCodecS = "";
	ieee64 fps = -1.0;
	int32 bitrate = -1;
	for (int i=3; i<argc; i++) {
		string s=argv[i];
		for (uint32 k=0; k<s.size(); k++)
		  s[k]=toupper(s[k]);
		
		if (s=="-MIRROR") mirror=true;
		else if (s=="-ROTATE") {
			i++;
			s=argv[i];
			rotate = atoi (s.c_str());
		} else if (s=="-CODEC") {
			i++;
			s=argv[i];
			for (uint32 k=0; k<s.size(); k++)
				s[k]=toupper(s[k]);
			encCodecS = s;
		} else if (s=="-FPS") {
			i++;
			s=argv[i];
			fps = atof (s.c_str());
		} else if (s=="-BITRATE") {
			i++;
			s=argv[i];
			bitrate = atoi (s.c_str());
		} else {
			cerr << "unknown option!" << endl;
			exit(1);
		} // else
	} // for


	// set destimation filename
	string dest = argv[2];


	// set rotation
	rotate /= 90;
	if (rotate < 0) rotate = 0;
	if (rotate > 3) rotate = 0;

	// set codec for encoder
	EncoderCodecs encCodec = CODEC_MPEG4;
	if (encCodecS != "") {
		if (encCodecS == "WMV1")
			encCodec = CODEC_WMV_1;

		else if (encCodecS == "WMV2")
			encCodec = CODEC_WMV_2;

		else if (encCodecS == "MPEG1")
			encCodec = CODEC_MPEG1;

		else if (encCodecS == "MPEG2")
			encCodec = CODEC_MPEG2;

		else if (encCodecS == "MPEG4")
			encCodec = CODEC_MPEG4;

		else {
			cerr << "unknown video codec!" << endl;
			exit(1);
		} // else
	} // if


	// set options
	encoderOpt.pathFilename = dest;
	encoderOpt.encCodec = encCodec;
	encoderOpt.silent=false;
	encoderOpt.framerate=fps;
	encoderOpt.bitrate=bitrate;

	return;
}

int main(int argc, char* argv[])
{

	cerr << "VideoConverter src dest [options]" << endl;
	cerr << "options:" << endl;
	cerr << "  -rotate 0,90,180,270                  rotate video to the right" << endl;
	cerr << "  -mirror                               mirror video" << endl;
	cerr << "  -codec  WMV1,WMV2,MPEG1,MPEG2,MPEG4   select encoder" << endl;
	cerr << "  -fps <number>                         set framerate, e.g. 25" << endl;
	cerr << "  -bitrate <number>                     valid bitrates: 4..16000 (in K)" << endl;

	cerr << endl;
	if (argc<3) {
		cerr << "name of source and destination video required!" << endl;
		exit(1);
	} // if

	string src  = argv[1];
	string dest = argv[2];


	int32 rotate;
	bool mirror;
	VideoEncoderOptions encoderOpt;
	parseParameters(argc, argv, encoderOpt, rotate, mirror);


	// open decoder
	VideoCodec imgCodec=CODEC_RGB;
	VideoDecoder reader (src, imgCodec);
	VideoSize size = reader.getSize();
	ImageSize imgSize = reader.getImageSize();

	// open encoder
	encoderOpt.imgCodec = imgCodec;
	encoderOpt.size = imgSize;
	VideoEncoder writer(encoderOpt);


	FlipFilter<uint8> flip (FLIP_HORIZONTAL);
	RotationFilter<uint8> rf (90.0*rotate);
	for (int32 i=0; i<size[DIM_Z]; i++) {

		Image<uint8> img = reader.getImage(i);
		cerr << "\rframe: " << i << "/" << size[DIM_Z] << "   " << ends;
		
		if (mirror) {
			flip.process(img);
		} // if
		if (rotate != 0) {
			rf.process(img);
		} // if

		writer.write(img);
	} // for
	cerr << "\r                                     " << endl;
	writer.close();
	cerr << "video convertion done!" << endl<<endl;

	return 0;
} // main

