/*
   This file is part of the MoCA video library.
   Name: AudioCutDetection.cpp
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: kopf@informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

/*! \file

   \author    Stephan Kopf

              Lehrstuhl fuer Praktische Informatik IV
              Universitaet Mannheim
              68131 Mannheim
              Germany
              email: kopf@informatik.uni-mannheim.de

   (c) Lehrstuhl fuer Praktische Informatik IV, details in COPYRIGHT.txt

*/

// C++ includes
#include <string>
#include <iostream>
using namespace std;

// MoCA includes
#include <FileList.h>
#include <AudioDecoder.h>
#include <WaveHeader.h>
//#include <IMpegAudioConvert.h>
#include <OWaveAudio.h>
#include <AudioEnergy.h>
#include <GaussFilter.h>
#include <AudioCut.h>


void usage()
{
	cerr << "Usage: AudioCutDetectionTest [options] filename|directory" << endl;
	cerr << "analyse the audio file (MPG,MP2,WAV) and show the audio cuts" << endl;
	cerr << endl;
	cerr << "Example:" <<  endl;
	cerr << "AudioCutDetectionTest test.wav" << endl;
	cerr << "Analyse the file \"test.wav\"" << endl;
	cerr << endl;
	cerr << "AudioCutDetectionTest -recurse D:/audio/" << endl;
	cerr << "Analyse all audio files in the  directory  \"D:/audio/\"." << endl;
	cerr << endl;
	exit(1);
} // usage


// blockLengthMs: default length of one audio block (samples during these milliseconds)
void analyseAudio (string &filename, Signal<uint8> &audioCuts, int32 blockLengthMilliSeconds=20)
{
	// initialize results
	SignalSize size (1);
	audioCuts.resize(size);
	audioCuts(0) = 0;

	// get file extension and convert it to lower case letters
	string fileExt="";
	FileHandling::getFileExtension (filename, fileExt);

	string srcMpg="";
	string srcMp2="";
	string srcWav="";
	if (fileExt == "mpg") {
		srcMpg = filename;
		srcMp2 = filename.substr(0, filename.size()-3);
		srcMp2 += "mp2";
		srcWav = filename.substr(0, filename.size()-3);
		srcWav += "wav";
	} // if
	if (fileExt == "mp2") {
		srcMp2 = filename;
		srcWav = filename.substr(0, filename.size()-3);
		srcWav += "wav";
	} // if
	if (fileExt == "wav") {
		srcWav = filename;
	} // if

	if (srcWav == "") {
		cerr << "AudioCutDetection: unknown audio file format (" << fileExt << ")!" << endl;
		return;
	} // if

	
	// extract WAV-file from MPG-file
	if (srcMpg != "") {
		try {
			// open reader
			AudioDecoderData data;
			AudioDecoder reader (srcMpg);
			reader.process (data);
		} catch (...) {
			cerr << endl << "AudioCutDetection: could not process audio (MPEG to WAV)." << endl;
			exit (-1);
		} // try
	} // if


	// there is a audio cut, if audioCutData.cuts(i)>5
	AudioCutData audioCutData;
	int32 minimumNumberOfCuts = 10;
	AudioCut audioCut(srcWav, minimumNumberOfCuts, blockLengthMilliSeconds);
	audioCut.process (audioCutData);


	

	return;
} // analyseAudio 


int main(int argc, char* argv[])
{
    if ((argc!=2) && (argc!=3)) usage();
	string filename =argv[argc-1];


    
	bool recurse=false;
	string s;
	for (int i=1; i<argc-1; i++) {
		s=argv[i];
		for (uint8 k=0; k<s.size(); k++)
		  s[k]=toupper(s[k]);
		
		if (s=="-RECURSE") {
			recurse=true;
		} else {
			if (i != argc-1) usage();
		} // else
	} // for
  

	Signal<uint8> audioCuts;
	if (!recurse) {
		// analyse audio
		analyseAudio(filename, audioCuts);
	} else {
		
		cerr << endl;
		cerr << "source path : " << filename << endl;
		
		// read all WAV files in directory
		vector<FileInfo> audioFiles;
		FileList files(filename, true);    
		files.process("*", audioFiles);
		
		// open mpeg files
		vector<FileInfo>::iterator it;
		for (it = audioFiles.begin(); it != audioFiles.end(); it++) {
			// analyse audio
			analyseAudio((*it).name, audioCuts);
		} // for
		cout << endl << audioFiles.size() << " files found." << endl;
	} // else
	
    return 0;
} // main


