/*
This file is part of the MoCA video library.
Name: CssMatchingTest.cpp
Copyright (C) 2004 Stephan Kopf

 Lehrstuhl fuer Praktische Informatik IV
 Universitaet Mannheim
 68131 Mannheim
 Germany
 email: kopf@informatik.uni-mannheim.de
 
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
	
*/

/*! \file

 
  
   \author    Stephan Kopf
   
	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de
	
	 (c) Lehrstuhl fuer Praktische Informatik IV, details in COPYRIGHT.txt
	 
	  $Id
	  
*/


// Moca includes
#include <Image.h>
#include <IImageStream.h>
#include <CssMatching.h>
#include <CssDisplayResult.h>


void usage()
{
	cerr << "Usage: CssMatching   database sequence [options]" << endl;
	cerr << "   database          path of database (tag files)" << endl;
	cerr << "   sequence          path of sequence (tag files)" << endl;
	cerr << "options:" << endl;
	cerr << "   -searchID=name    print only the results of \"name\", e.g. \"Cars\" " << endl;
	cerr << "   -width=false      deactivate the width as matching factor" << endl;
	cerr << "   -ratio=false      deactivate the aspect ratio as matching factor" << endl;
	cerr << "   -store-matrix     store a matrix with all the best match for each object class" << endl;
	cerr << "   -rotateCSS=false  do not rotate the css image" << endl;
	cerr << "   -global           use global shape descriptors" << endl;
	cerr << "   -extended         use extended css matching (mirror contours)" << endl;
	cerr << endl;
	cerr << "Example:" <<  endl;
	cerr << "CssMatching  D:/css/database/ D:/css/sequence/changick/  -searchID=People -width=false -ratio=false" << endl;
	exit(1);
} // usage




int main(int argc, char* argv[])
{
	
	// set parameters (paths)
	string databasePath;			// path with database tag files
	string sequencePath;			// path with sequence tag files
    string fileExtension = "*CSS.tag";
	string searchId = "";
	
	
	if (argc<3) usage();
	databasePath = argv [1];
    sequencePath = argv [2];
	
	
	bool useWidth=true;
	bool useRatio=true;
	bool rotateCss=true;
	bool storeMatrix=false;
	bool extended=false;
	bool globalDescriptors=false;
	for (int i=3; i<argc; i++) {
		string s=argv[i];
		for (uint32 k=0; k<s.size(); k++)
			s[k]=toupper(s[k]);
		
		if (s=="-WIDTH=FALSE") useWidth=false;
		else {
			if (s=="-WIDTH=TRUE") useWidth=true;
			else {
				if (s=="-ROTATECSS=FALSE") rotateCss=false;
				else {
					if (s=="-RATIO=FALSE") useRatio=false;
					else {
						if (s=="-RATIO=TRUE") useRatio=true;
						else {
							if (s=="-EXTENDED") extended=true;
							else {
								if (s=="-GLOBAL") globalDescriptors=true;
								else {
									if (s=="-STORE-MATRIX") storeMatrix=true;
									else {
										string t=s.substr (0,10);
										if (t=="-SEARCHID=") searchId = s.substr (10);
										else usage();
		}}}}}}}}
	} // for
	
	string localPath="";
	if (storeMatrix) {
		FileHandling::getCurrentDirectory(localPath);
	} // if
	
	
	// open database
    cout << "read database files " << endl;
    cout << databasePath << flush;
    vector<FileInfo> databaseList;
    FileList flData (databasePath, true);
    flData.process (fileExtension, databaseList);
    cout << "("<< databaseList.size() << ")"<< endl;
	
	
	// open sequence (unknown images)
    cout << "read sequence files " << endl;
    cout << sequencePath << flush;
    vector<FileInfo> sequenceList;
    FileList flSequence (sequencePath, true);
    flSequence.process (fileExtension, sequenceList);
    cout << "("<< sequenceList.size() << ")"<< endl;
    
#if 1
	globalDescriptors=true;
#endif
	
	
	// initialize databases 
	CssDisplayResult result (databaseList, sequenceList, false, true, globalDescriptors);    
	
	
	int tPosition;
	int tHeight;
	int tWidth;
	int tCriticalHeight;
	bool tCheckConcave;
	bool XYratio=true;
	
	
	// define default parameters
	tCriticalHeight = 50; //50; 
	tPosition = 30; //30; 
	tHeight = 30; //50; //30; 
	tWidth = 60; // 85; //60; 
	
	tCheckConcave = false;
	XYratio = false;
	rotateCss = true;
	//extended = false;
	extended = true;
	storeMatrix = false;
	//globalDescriptors = false;
	globalDescriptors = true;
	
	
	// matching with parameters (e.g. position=20%, height=20%, width=40%, criticalHeight=40%, checkConcave=false)
	//XYratio=true;
	CssMatching cssm(
		ContourParametrization::DEFAULT_CONTOUR_SIZE, 
		tPosition,
		tHeight,
		tWidth,
		tCriticalHeight,
		tCheckConcave,
		XYratio,
		rotateCss,
		globalDescriptors); 
	
	// calculate distances
	int32 totalRecjected=0;
	int i, j;    
	if (globalDescriptors) {
		cerr << "use global descriptors" << endl;
	} else {
		cerr << "use NO global descriptors" << endl;
	} // else
	if (extended) {
		cerr << "use extended css appraoch" << endl;
	} else {
		cerr << "use standard css appraoch, only" << endl;
	} // else
	for (i=0; i<result.getSequenceSize(); i++) {
		for (j=0; j<result.getDatabaseSize(); j++) {								
			CssDifference d;
			d.distance = 0;
			
			if (globalDescriptors) {
				ieee64 seqC  = result.seq[i].geometric.compactness;
				ieee64 seqE  = result.seq[i].geometric.eccentricity;
				ieee64 dataC = result.data[j].geometric.compactness;
				ieee64 dataE = result.data[j].geometric.eccentricity;
				
				ieee64 fC = max(seqC, dataC) / min(seqC, dataC);
				ieee64 fE = max(seqE, dataE) / min(seqE, dataE);
				
				const ieee64 thC=1.5; //1.5;
				const ieee64 thE=1.3; //1.3;
				//cerr << "TH: " << thC << "   " << thE << "  " << endl;
				
				if ((fC > thC) || (fE > thE)) {
					totalRecjected++;
					d.distance = INT_MAX;
				} // if 
			} // if
			
			
			if (d.distance == 0) {
				if (extended) {
					d.distance = cssm.processExtended(
						result.data[j].cmS, result.data[j].cmX, 
						result.seq[i].cmS,  result.seq[i].cmX  );
				} else {
					d.distance = cssm.process(result.data[j].cmS, result.seq[i].cmS);
				} // else
			} // if
			d.dbRef = j;
			result.seq[i].diff.push_back(d);
		} // for
	} // for
	cout << "total rejected: " << totalRecjected << endl;
	cerr << "total rejected: " << totalRecjected << endl;
	
	if (searchId == "") {
		result.showAll(6);
	} else {
		result.show(searchId);
	} // else
	
	// store a matrix:
	// each line (=object class) stores the best values for each frame
	if (storeMatrix) {
		string matrixFile=localPath + "cssMatrix.txt";
		result.printMatrix(matrixFile);
	} // if
	
	return 0;
} // main











