/*
   This file is part of the MoCA video library.
   Name: GeometricFeatureMatchingTest.cpp
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: kopf@informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


/*! \file



   \author    Stephan Kopf

              Lehrstuhl fuer Praktische Informatik IV
              Universitaet Mannheim
              68131 Mannheim
              Germany
              email: kopf@informatik.uni-mannheim.de

   (c) Lehrstuhl fuer Praktische Informatik IV, details in COPYRIGHT.txt

   $Id

*/


// Moca includes
#include <Image.h>
#include <IImageStream.h>
#include <FileList.h>
#include <GeometricFeature.h>


void usage()
{
	cerr << "Usage: GeometricMatching   database sequence [options]" << endl;
	cerr << "   database          path of database (tag files)" << endl;
	cerr << "   sequence          path of sequence (tag files)" << endl;
	cerr << endl;
	cerr << "Example:" <<  endl;
	cerr << "GeometricMatching  D:/css/database/ D:/css/sequence/changick/" << endl;
	exit(1);
} // usage


struct MatchingData
{
	MatchingData() {
		fn = "";
		objClass="";
		objClassID=0;
	}

	GeometricFeatureData geometric;
	string fn;
	string objClass;
	int32 objClassID;
};


void match (vector<MatchingData> &database, vector<MatchingData> &sequence)
{


	//const ieee64 thCompactness  = 1.50;
	//const ieee64 thEccentricity = 1.33;
	const ieee64 thCompactness  = 1.25; //1.7;
	const ieee64 thEccentricity = 1.15; //1.5;
	//const ieee64 thBoth         = 103.10;

	const int32 MAX_OBJECT_CLASSES=255;
	int32 similarObjectClass [MAX_OBJECT_CLASSES];
	for (int32 i=0; i<MAX_OBJECT_CLASSES; i++) {
		similarObjectClass[i]=0;
	} // for
	int32 totalObjectClass [MAX_OBJECT_CLASSES];
	for (int32 i=0; i<MAX_OBJECT_CLASSES; i++) {
		totalObjectClass[i]=0;
	} // for

	int32 cnt=0;
	for (int32 i=0; i<sequence.size(); i++) {
		ieee64 srcCompactness = sequence[i].geometric.compactness;
		ieee64 srcEccentricity= sequence[i].geometric.eccentricity;

		// match with database
		for (int32 j=0; j<database.size(); j++) {
			ieee64 dbCompactness = database[j].geometric.compactness;
			ieee64 dbEccentricity= database[j].geometric.eccentricity;

			ieee64 fCompactness = max(srcCompactness, dbCompactness ) / min(srcCompactness, dbCompactness );
			ieee64 fEccentricity= max(srcEccentricity,dbEccentricity) / min(srcEccentricity,dbEccentricity);

			bool similar=true;
			if (fCompactness > thCompactness ) similar=false;
			if (fEccentricity> thEccentricity) similar=false;
			//if (fCompactness+fEccentricity > thBoth) similar=false;
			
			if (similar) {
				similarObjectClass[database[j].objClassID]++;
			} // if
			totalObjectClass[database[j].objClassID]++;
		} // for
	} // for

	// print results
	cerr << "percentage of the most similar object classes:" << endl;
	int32 v= min(MAX_OBJECT_CLASSES,7+1);
	ieee64 sum=0;
	ieee64 sum2=0;
	for (int32 i=1; i<v; i++) {
		ieee64 g = similarObjectClass[i];
		sum+=g;
		sum2+=totalObjectClass[i];
		if (totalObjectClass[i] > 0) g/=totalObjectClass[i];
		cout << (int32)(100.0*g) << "\t";
		//cerr << similarObjectClass[i] << endl;
	} // for
	cout << endl;

	
	cout << "-----------------" << endl;
	cout << sequence.size() << "\t" << ends;
	for (int32 i=1; i<v; i++) {
		ieee64 hh=totalObjectClass[i]-similarObjectClass[i];
		hh /= sequence.size();
		cout << (int)(hh) << "\t";
	} // for
	cout << endl;
	cout << "-----------------" << endl;

	cout << "total: " << (int32)(100.0*sum/sum2) << endl;
	exit(1);

	return;
}


void getAverageDatabase (vector<MatchingData> &database)
{
	// get average
	ieee64 avgCompactness = 0;
	ieee64 avgEccentricity  = 0;
	ieee64 n=0;
	for (int32 i=0; i<database.size(); i++) {
		MatchingData d = database[i];
		avgCompactness  += d.geometric.compactness;
		avgEccentricity += d.geometric.eccentricity;
		n++;
	} // for
	if (n<=0) n=1;
	avgCompactness /=n;
	avgEccentricity/=n;
	ieee64 sigCompactness = 0;
	ieee64 sigEccentricity  = 0;
	for (int32 i=0; i<database.size(); i++) {
		MatchingData d = database[i];
		sigCompactness  += (avgCompactness - d.geometric.compactness) * (avgCompactness - d.geometric.compactness);
		sigEccentricity += (avgEccentricity -d.geometric.eccentricity)* (avgEccentricity -d.geometric.eccentricity);
	} // for
	sigCompactness /=n;
	sigEccentricity/=n;
	if (sigCompactness >0) sigCompactness =sqrt(sigCompactness);
	if (sigEccentricity>0) sigEccentricity=sqrt(sigEccentricity);


	cerr << endl << "DB: " << endl;
	cerr << "compactness : " << avgCompactness  << "   \t" << sigCompactness << endl;
	cerr << "eccentricity: " << avgEccentricity << "   \t" << sigEccentricity << endl;

	return;
}


#include<Video.h>
//#include<OVideoStream.h>
//#include<OPgmVideoFormatStream.h>
//#include<fstream>
#include <PpmVideoDecoder.h>
#include <PpmVideoEncoder.h>
#include <CropFilter.h>
#include <DrawLine.h>





int main(int argc, char* argv[])
{


	// set parameters (paths)
	string databasePath;			// path with database tag files
	string sequencePath;			// path with sequence tag files
    string fileExtension = "*_GEOMETRIC.tag";

	if (argc<3) usage();
	if (argc>3) usage();
	databasePath = argv [1];
    sequencePath = argv [2];


	// read data
	vector<MatchingData> database;
	{
		// open database
		cout << "read database files " << endl;
		cout << databasePath << flush;
		vector<FileInfo> databaseList;
		FileList flData (databasePath, true);
		flData.process (fileExtension, databaseList);
		cout << " ("<< databaseList.size() << ")"<< endl;
		
		for (int32 i=0; i<databaseList.size(); i++) {
			string fn = databaseList[i].name;
			MatchingData data;
			data.fn = fn;

			// set object class
			int32 p = fn.find_last_of('/');
			fn = fn.substr(0,p);
			p = fn.find_last_of('/');
			fn = fn.substr(0,p);
			p = fn.find_last_of('/');
			fn = fn.substr(0,p);
			p = fn.find_last_of('/');
			data.objClass = fn.substr(p+1);
			FileHandling::upper(data.objClass);
			
			// set object class ID
			char c = data.objClass[0];
			data.objClassID = (c - '0');

			// read data
			ITagStream its (data.fn);
			its >> data.geometric;		
			its.close();
			
			database.push_back(data);
		} // for
	}

	// read data
	vector<MatchingData> sequence;
	{
		// open sequence (unknown images)
		cout << "read sequence files " << endl;
		cout << sequencePath << flush;
		vector<FileInfo> sequenceList;
		FileList flSequence (sequencePath, true);
		flSequence.process (fileExtension, sequenceList);
		cout << " ("<< sequenceList.size() << ")"<< endl;
		
		for (int32 i=0; i<sequenceList.size(); i++) {
			string fn = sequenceList[i].name;
			MatchingData data;
			data.fn = fn;
			
			// read data
			ITagStream its (data.fn);
			its >> data.geometric;		
			its.close();
			
			sequence.push_back(data);
		} // for
	}

	// get average values
	//getAverageDatabase (database);

	
	// match data
	match (database, sequence);

	return 0;
} // main



