/*
    ActionDetection.cpp     detects action in videos
    Copyright (C) 2004 Stephan Kopf

	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// MoCA includes
#include <ActionDetection.h>


std::ostream& operator<< (ostream &os, const ActionData &data) 
{
	os << "version               : " << data.version << endl;
	os << "filename              : " << data.filename << endl;
	os << "number of frames      : " << data.numberOfFrames << endl;
	os << "average image action  : " << data.avgImgAction << endl;
	os << "average motion action : " << data.avgMotionAction << endl;
	os << endl;

	/*
	os << "frame \timage difference \tactiondifference" << endl;
	for (int32 i=0; i<data.numberOfFrames; i++) {
		os << setw(6) << i << "\t" 
		   << setw(6) << data.imgAction(i) << "\t" 
		   << setw(6) << data.motionAction(i) << endl;
	} // for
	*/

	return os;
} // operator <<

OTagStream& operator<< (OTagStream& os, const ActionData &data) 
{
    TagFormat::writeStartTag (os, "ACTION");
	os << endl;
	
    TagFormat::writeNodeValue(os, "version", data.version, 1);
    TagFormat::writeNodeValue(os, "filename", data.filename, 1);
    TagFormat::writeNodeValue(os, "numberOfFrames", data.numberOfFrames, 1);
    TagFormat::writeNodeValue(os, "avgImageAction", data.avgImgAction, 1);
    TagFormat::writeNodeValue(os, "avgMotionAction", data.avgMotionAction, 1);
	os << endl;
	
	// write image difference data
	TagFormat::writeStartTag (os, "IMAGE_ACTION", 1);
	os << endl;
	for (int32 i=0; i<data.imgAction.getSize()[DIM_X]; i++) {
		os << "     " << data.imgAction(i) << endl; 
	} // for
	TagFormat::writeEndTag (os, "IMAGE_ACTION", 1);
	os << endl;

	
	// write motion data
	TagFormat::writeStartTag (os, "MOTION_ACTION", 1);
	os << endl;
	for (int32 i=0; i<data.motionAction.getSize()[DIM_X]; i++) {
		os << "     " << data.motionAction(i) << endl; 
	} // for
	TagFormat::writeEndTag (os, "MOTION_ACTION", 1);
	os << endl;

    TagFormat::writeEndTag (os, "ACTION");
	os << endl;
	
    return os;
} // operator<<


ITagStream& operator>> (ITagStream& is, ActionData& data)
{
	
    char c;
    bool failFlag = true;
	
	if (is >> c) {
		failFlag = false;
		is.putback(c);

		TagFormat::readStartTag (is, "ACTION");
		
		TagFormat::readNodeValue(is, "version", data.version);
		TagFormat::readNodeValue(is, "filename", data.filename);
		TagFormat::readNodeValue(is, "numberOfFrames", data.numberOfFrames);
		TagFormat::readNodeValue(is, "avgImageAction", data.avgImgAction);
		TagFormat::readNodeValue(is, "avgMotionAction", data.avgMotionAction);
		if (is.rdstate() & ios::failbit) {
			throw MocaException ("ActionDetection: Illegal index file format!");
		} // if

		// initialize 
		data.imgAction.resize(SignalSize(data.numberOfFrames));
		data.motionAction.resize(SignalSize(data.numberOfFrames));
		for (int32 i=0; i<data.numberOfFrames; i++) {
			data.imgAction(i) = 0.0;
			data.motionAction(i) = 0.0;
		} // for

		// read cuts
		if (is >> c) {
			is.putback(c);

			// read image action
			TagFormat::readStartTag(is, "IMAGE_ACTION");
			for (int32 i=0; i<data.numberOfFrames; i++) {
				ieee64 p=0.0;
				is >> p;
				data.imgAction(i) = p;
			} // for
			TagFormat::readEndTag(is, "IMAGE_ACTION");
			if (is.rdstate() & ios::failbit) {
				throw MocaException ("ActionDetection: Illegal index file format!");
			} // if

			// read motion action
			TagFormat::readStartTag(is, "MOTION_ACTION");
			for (int32 i=0; i<data.numberOfFrames; i++) {
				ieee64 p=0.0;
				is >> p;
				data.motionAction(i) = p;
			} // for
			TagFormat::readEndTag(is, "MOTION_ACTION");
			if (is.rdstate() & ios::failbit) {
				throw MocaException ("ActionDetection: Illegal index file format!");
			} // if
		} // if
		
		TagFormat::readEndTag (is, "ACTION");
	} // if

	if (is.rdstate() & ios::failbit) {
		throw MocaException ("ActionDetection: Illegal index file format!");
	} else {
        is.clear();
    } // else
	
    return is;
} // operator>> 





ActionDetection::ActionDetection(const string &pathFilename) 
{
	this->pathFilename = pathFilename;
} 


void 
ActionDetection::process(ActionData& data) 
{
	// read data
	ReadMocaData<ActionData> reader (pathFilename, ACTION_DETECTION_FILENAME);
	bool dataValid = reader.process (ACTION_DETECTION_VERSION, data); 
	
	if (!dataValid) {
		
		// create new index
		cerr << "\rActionDetection: Creating index file ...          " << endl;
		readValues(data);
		
		// save index file
		WriteMocaData<ActionData> writer (pathFilename, ACTION_DETECTION_FILENAME);
		writer.process(data);
		cerr << "\rActionDetection: Creating index file ... done          " << endl;
	} // if

} // process


void 
ActionDetection::readValues (ActionData& data) 
{
	// set variables
	data.version = ACTION_DETECTION_VERSION;
	data.numberOfFrames = 0;
	data.avgImgAction = 0.0;
	data.avgMotionAction = 0.0;


	// get path of filename, do not create new directory
	CreateMocaPath mocaPath (pathFilename, false);
	data.filename = mocaPath.file;

	// resize signal
	VideoCodec codec = CODEC_Y;

	VideoDecoder reader (pathFilename, 3, -1, codec);
	VideoSize vidSize = reader.getSize();
	data.numberOfFrames = vidSize[DIM_Z];
	data.imgAction.resize(SignalSize(data.numberOfFrames));
	data.motionAction.resize(SignalSize(data.numberOfFrames));
	for (int32 i=0; i<data.numberOfFrames; i++) {
		data.imgAction(i) = 0.0;
		data.motionAction(i) = 0.0;
	} // for


	// analyse motion action
	MotionVectorStatus motion;
	int32 searchRange = 32;
	MotionVector mv (pathFilename, searchRange);
	mv.process (motion);

	// guess: the largest motion vector should have a length of 16.0
	//        if the average motion vector exceeds 16, the motion action is set to 1.0
	ieee64 numberOfBlocks= 1.0/(16.0*motion.size[DIM_X]*motion.size[DIM_Y]);
	Image<ieee64> motionVectors;
	for (int32 i=0; i<motion.size[DIM_Z]; i++) {
		cerr << "\rcalc mv: " << i << ends;
		ieee64 sum=0.0;

		mv.getMotionVector(i, motionVectors);
		for (int32 y=0; y<motion.size[DIM_Y]; y++) {
			for (int32 x=0; x<motion.size[DIM_X]; x++) {
				ieee64 v0=motionVectors(x,y,0);
				ieee64 v1=motionVectors(x,y,1);
				sum+=sqrt(v0*v0+v1*v1);
			} // for
		} // for
		sum*=numberOfBlocks;
		if (sum>1.0) sum=1.0;
		data.motionAction(i)=sum;
		data.avgMotionAction+=sum;
	} // for
	data.avgMotionAction/=((ieee64)motion.size[DIM_Z]);


	// analyse image action
	Image<uint8> img[2];
	bool nextImg=0;
	img[nextImg]=reader.getImage(0);
	nextImg=(nextImg+1)%2;
	cerr << "\r                        \r"<< ends;
	for (int32 i=1; i<data.numberOfFrames; i++) {
		img[nextImg]=reader.getImage(i);
		cerr << "\rframe: "<< i << "   " << ends;
		nextImg=(nextImg+1)%2;
		int32 diff=0;
		for (int32 y=0; y<vidSize[DIM_Y]; y++) {
			for (int32 x=0; x<vidSize[DIM_X]; x++) {
				diff+=labs ((int32)img[0](x,y) - (int32)img[1](x,y));
			} // for
		} // for
		ieee64 diff2=(diff / (255.0*vidSize[DIM_X]*vidSize[DIM_Y]));
		data.imgAction(i-1)=diff2;
		data.avgImgAction+=diff2;
	} // for
	data.avgImgAction/=((ieee64)data.numberOfFrames-1);

	return;	
} // readValues
