/*
    CameraTransformation.cpp     get camera transformation between two frames
    Copyright (C) 2004 Stephan Kopf

	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// MoCA includes
#include <CameraTransformation.h>


std::ostream& operator<< (ostream& os, const TransformationData &item)
{	
	os << item.a00 << " ";
	os << item.a01 << " ";
	os << item.a10 << " ";
	os << item.a11 << " ";

	os << item.tx << " ";
	os << item.ty << " ";
	os << item.px << " ";
	os << item.py << " ";

	os << endl;
	
	return os;
} // operator<<

OTagStream& operator<< (OTagStream& os, const TransformationData& item)
{
	TagFormat::writeStartTag(os, "parameters");
	os << item.a00 << " " << item.a01 << " " << item.a10 << " " << item.a11 << " " 
	   << item.tx  << " " << item.ty  << " " << item.px  << " " << item.py << " ";
	TagFormat::writeEndTag(os, "parameters");

    return os;
} // operator<<

ITagStream& operator>> (ITagStream& is, TransformationData& item)
{
    char c;
    bool failFlag = false;
	
	if (is >> c) {
		is.putback(c);
		
		TagFormat::readStartTag(is, "parameters");
		is >> item.a00;
		is >> item.a01;
		is >> item.a10;
		is >> item.a11;
		is >> item.tx;
		is >> item.ty;
		is >> item.px;
		is >> item.py;
		TagFormat::readEndTag(is, "parameters");
	} // if
		
	if (is.rdstate() & ios::failbit) {
		failFlag = true;
	} // if
	if (failFlag) {
		throw MocaException ("CameraTransformation: Illegal file format!");
	} else {
		is.clear();
	} // else
	
	return is;
} // operator>> 



std::ostream& operator<< (ostream& os, const list<TransformationData> &data)
{
	int32 n = data.size();
	os << "number of items: " << n << endl;
	
	list<TransformationData>::const_iterator it;
	for (it = data.begin(); it!= data.end(); it++) {
		os << (*it) << endl;
	} // for
	
	return os;
} // operator<<


OTagStream& operator<< (OTagStream& os, const list<TransformationData>& data)
{
	int32 n = data.size();

	TagFormat::writeStartTag(os, "CAMERA_TRANSFORMATION");
	os << endl;
    TagFormat::writeNodeValue(os, "numberOfItems", n);
	os << endl;


	list<TransformationData>::const_iterator it;
	for (it = data.begin(); it!= data.end(); it++) {
		os << (*it);
	} // for
	os << endl;

	TagFormat::writeEndTag(os, "CAMERA_TRANSFORMATION");

	return os;
} // operator<<

ITagStream& operator>> (ITagStream& is, list<TransformationData>& item)
{
	item.clear();
    char c;
    bool failFlag = false;
	
	if (is >> c) {
		is.putback(c);
		

		TagFormat::readStartTag(is, "CAMERA_TRANSFORMATION");

		int32 n=0;
		TagFormat::readNodeValue(is, "numberOfItems", n);

		for (int32 i=0; i<n; i++) {
			TransformationData tmp;
			is >> tmp;
			item.push_back(tmp);
		} // for
		TagFormat::readEndTag(is, "CAMERA_TRANSFORMATION");
	} // if
		
	if (is.rdstate() & ios::failbit) {
		failFlag = true;
	} // if
	if (failFlag) {
		throw MocaException ("CameraTransformation: Illegal file format!");
	} else {
		is.clear();
	} // else
	
	return is;
} // operator>> 





void
TransformationData::processInverse (const TransformationData &src, TransformationData &dest)
{
	TransformationPoint<double> start[4];
	TransformationPoint<double> end[4];
	
	for (int i=0;i<4;i++) {
		start[i].x = (i%2) ? 1.0 : 0.0;
		start[i].y = (i<2) ? 0.0 : 1.0;
		processPoint(src, start[i], end[i]);
    } // for
	
	PerspFrom4PointCorrespondences(end,start, dest);
	return;
}



void 
TransformationData::process (TransformationData& trans, TransformationData& oldNewTrans)
{
	TransformationPoint<double> start[4];
	TransformationPoint<double> end[4];
	TransformationPoint<double> tmp;
	
	for (int i=0;i<4;i++) {
		start[i].x = (i%2) ? 1.0 : 0.0;
		start[i].y = (i<2) ? 0.0 : 1.0;
		processPoint(trans, start[i], tmp);
		processPoint(oldNewTrans, tmp, end[i]);
    } // for

	PerspFrom4PointCorrespondences(start,end, oldNewTrans);
	return; 
}


void
TransformationData::processPoint(const TransformationData &trans, const TransformationPoint<double>& src, TransformationPoint<double> &dest) 
{
	dest.x = (src.x*trans.a00 + src.y*trans.a01 + trans.tx)/(src.x*trans.px + src.y*trans.py + 1.0);
	dest.y = (src.x*trans.a10 + src.y*trans.a11 + trans.ty)/(src.x*trans.px + src.y*trans.py + 1.0);
	return;
}


void
TransformationData::PerspFrom4PointCorrespondences (const TransformationPoint<double>* start, const TransformationPoint<double>* end, TransformationData &trans)
{
	gsl_vector* rhs      = gsl_vector_alloc(8);
	gsl_matrix* matrix   = gsl_matrix_alloc(8,8);
	gsl_permutation* per = gsl_permutation_calloc(8);
	
	for (int i=0;i<4;i++) {
		gsl_matrix_set(matrix,2*i  ,0,start[i].x);
		gsl_matrix_set(matrix,2*i  ,1,start[i].y);
		gsl_matrix_set(matrix,2*i  ,2,1         );
		gsl_matrix_set(matrix,2*i  ,3,0         );
		gsl_matrix_set(matrix,2*i  ,4,0         );
		gsl_matrix_set(matrix,2*i  ,5,0         );
		gsl_matrix_set(matrix,2*i  ,6,-start[i].x * end[i].x);
		gsl_matrix_set(matrix,2*i  ,7,-start[i].y * end[i].x);
		
		gsl_vector_set(rhs,2*i  ,end[i].x);
		
		gsl_matrix_set(matrix,2*i+1,0,0         );
		gsl_matrix_set(matrix,2*i+1,1,0         );
		gsl_matrix_set(matrix,2*i+1,2,0         );
		gsl_matrix_set(matrix,2*i+1,3,start[i].x);
		gsl_matrix_set(matrix,2*i+1,4,start[i].y);
		gsl_matrix_set(matrix,2*i+1,5,1         );
		gsl_matrix_set(matrix,2*i+1,6,-start[i].x * end[i].y);
		gsl_matrix_set(matrix,2*i+1,7,-start[i].y * end[i].y);
		
		gsl_vector_set(rhs,2*i+1,end[i].y);
    } // for
	
	int signum=0;
	gsl_linalg_LU_decomp(matrix,per,&signum);
	gsl_linalg_LU_svx(matrix,per,rhs);

	
	trans.a00 = gsl_vector_get(rhs,0);
	trans.a01 = gsl_vector_get(rhs,1);
	trans.tx      = gsl_vector_get(rhs,2);
	trans.a10 = gsl_vector_get(rhs,3);
	trans.a11 = gsl_vector_get(rhs,4);
	trans.ty      = gsl_vector_get(rhs,5);
	trans.px      = gsl_vector_get(rhs,6);
	trans.py      = gsl_vector_get(rhs,7);
	
	gsl_vector_free(rhs);
	gsl_permutation_free(per);
	gsl_matrix_free(matrix);
	
	return;
}

// **********************



void 
PerspectiveTransformation::TransformBitmap_LinInterpol (
	const Image<uint8>& src, Image<uint8>& dest, ImageSize &destSize,
	TransformationData &trans, uint8 background,  
	int32 dest_left,int32 dest_top, int32 dest_right,int32 dest_bottom)
{
	if (destSize[DIM_BAND] != 1) destSize[DIM_BAND] = 1;
	dest.resize (destSize);
	dest.clear(background); // set dest to background color
	
	if (dest_left < 0) dest_left = 0;
	if (dest_top < 0) dest_top = 0;
	dest_right  = min(destSize[DIM_X]-1,dest_right);
	dest_bottom = min(destSize[DIM_Y]-1,dest_bottom);
	
	int32 srcWidth  = src.getSize()[DIM_X];
	int32 srcHeight = src.getSize()[DIM_Y];

	TransformationPoint<double> destP;
	TransformationPoint<double> p;
	for (int32 x=dest_left;x<=dest_right;x++) {
		for (int32 y=dest_top;y<=dest_bottom;y++)
		{
			destP.x = x;
			destP.y = y;
			
			TransformationData::processPoint(trans, destP, p);
			int32 ix = (int32)floor(p.x);
			int32 iy = (int32)floor(p.y);
			double wx = p.x-ix;
			double wy = p.y-iy;
			
			if ( (ix<srcWidth) && (iy<srcHeight) && (ix>=0) && (iy>=0) ) {
				dest(x,y) = (uint8) 
					(
					(src(ix,iy)  *(1.0-wx) + src(ix+1,iy)  *(wx))  * (1.0-wy) +
					(src(ix,iy+1)*(1.0-wx) + src(ix+1,iy+1)*(wx))  *   (  wy)
					);
			} // if
		} // for
	} // for
	return;
}



void 
PerspectiveTransformation::TransformBitmap_NonLinInterpol (
	const Image<uint8>& src, Image<uint8>& dest, ImageSize &destSize,
	TransformationData &trans, uint8 background,  
	int32 dest_left,int32 dest_top, int32 dest_right,int32 dest_bottom)
{
	if (destSize[DIM_BAND] != 1) destSize[DIM_BAND] = 1;
	dest.resize (destSize);
	dest.clear(background); // set dest to background color
	
	if (dest_left < 0) dest_left = 0;
	if (dest_top < 0) dest_top = 0;
	dest_right  = min(destSize[DIM_X]-1,dest_right);
	dest_bottom = min(destSize[DIM_Y]-1,dest_bottom);
	
	int32 srcWidth  = src.getSize()[DIM_X];
	int32 srcHeight = src.getSize()[DIM_Y];

	TransformationPoint<double> destP;
	TransformationPoint<double> p;
	for (int32 x=dest_left;x<=dest_right;x++) {
		for (int32 y=dest_top;y<=dest_bottom;y++)
		{
			destP.x = x;
			destP.y = y;
			TransformationData::processPoint(trans, destP, p);

			int32 ix = (int32)floor(p.x);
			int32 iy = (int32)floor(p.y);
			double wx = p.x-ix;
			double wy = p.y-iy;

			if (wx>=0.5) ix++;
			if (wy>=0.5) iy++;
			if ( (ix<srcWidth) && (iy<srcHeight) && (ix>=0) && (iy>=0) ) {
				dest(x,y) = src(ix,iy);
			} // if
		} // for
	} // for
	return;
}


