/*
   This file is part of the MoCA video library.
   Name: CameraTransformation.h
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: kopf@informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

/*! \file

 
  
   \author    Stephan Kopf
   
	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de
	
	 (c) Lehrstuhl fuer Praktische Informatik IV, details in COPYRIGHT.txt
	 
*/


#ifndef CAMERA_TRANSFORMATION_H
#define CAMERA_TRANSFORMATION_H

// C includes
#include <math.h>

// C++ includes
#include <iostream>
using namespace std;

// MoCA includes
#include <MocaTypes.h>
#include <Image.h>
#include <ConvertColorSpace.h>



// GSL includes
// be sure that the GSL-libraries is included in the project (lib\gsl.lib lib\gslcblas.lib)
#include <gsl/gsl_linalg.h>



template <class T> struct TransformationPoint
{
    TransformationPoint() {}
    TransformationPoint(const T& _x,const T& _y) 
	{ 
		x = _x; 
		y = _y; 
	}
	
    T x,y;
};

struct TransformationPointInt32
{
    TransformationPointInt32() {}
    TransformationPointInt32(const int32& _x,const int32& _y) 
	{ 
		x = _x; 
		y = _y; 
	}
	
    int32 x,y;
};

struct TransformationData
{
	TransformationData() 
	{
		a00=a11=1.0; 
		a01=a10=0.0; 
		tx=ty=px=py=0.0; 
	}

	TransformationData(ieee64 a00, ieee64 a01, ieee64 a10, ieee64 a11, 
					   ieee64 tx, ieee64 ty, ieee64 px, ieee64 py) 
	{
		this->a00=a00;
		this->a11=a11; 
		this->a10=a10; 
		this->a01=a01; 

		this->tx=tx; 
		this->ty=ty; 
		this->px=px; 
		this->py=py; 
	}


   TransformationData& operator=(const TransformationData &other) {
       if (this != &other) {
           a00 = other.a00;
           a01 = other.a01;
           a10 = other.a10;
           a11 = other.a11;
           tx      = other.tx;
           ty      = other.ty;
           px      = other.px;
           py      = other.py;
       } // if
       return *this;
   } // operator=

   void clear()
   {
		a00=a11=1.0; 
		a01=a10=0.0; 
		tx=ty=px=py=0.0; 
		return;
   }

	// apply transformation "trans" to existing transformation 
	static void process (TransformationData& trans, TransformationData& oldNewTrans);

	// the inverse transformation is calculated and stored in "dest"
	static void processInverse (const TransformationData &src, TransformationData &dest);

	// the point "src" is transformed to the point "dest"
	static void processPoint(const TransformationData &trans, const TransformationPoint<double>& src, TransformationPoint<double> &dest);

	// get the maximum length of a motion vector (based on image size)
	static void getMotionVectorLength (const TransformationData &trans, ImageSize size, ieee64 &length)
	{
		ieee64 dx = size[DIM_X]-1;
		ieee64 dy = size[DIM_Y]-1;
		TransformationPoint<double> dest;
		length=0.0;
		{
			TransformationPoint<double> src (0,0);
			processPoint(trans, src, dest);
			ieee64 v = (dest.x-src.x) * (dest.x-src.x) + (dest.y-src.y) * (dest.y-src.y);
			if (length<v) length=v;
		}
		{
			TransformationPoint<double> src (dx,0);
			processPoint(trans, src, dest);
			ieee64 v = (dest.x-src.x) * (dest.x-src.x) + (dest.y-src.y) * (dest.y-src.y);
			if (length<v) length=v;
		}
		{
			TransformationPoint<double> src (dx,dy);
			processPoint(trans, src, dest);
			ieee64 v = (dest.x-src.x) * (dest.x-src.x) + (dest.y-src.y) * (dest.y-src.y);
			if (length<v) length=v;
		}
		{
			TransformationPoint<double> src (0,dy);
			processPoint(trans, src, dest);
			ieee64 v = (dest.x-src.x) * (dest.x-src.x) + (dest.y-src.y) * (dest.y-src.y);
			if (length<v) length=v;
		}
		if (length>0) length = sqrt(length);
		return;
	}

	// calculate transformation
	static void PerspFrom4PointCorrespondences(const TransformationPoint<double>* start, const TransformationPoint<double>* end, TransformationData &trans);
private:


public:
	ieee64 a00,a01,a10,a11;
	ieee64 tx,ty;
	ieee64 px,py;
};

std::ostream& operator<< (ostream&,    const TransformationData&);
OTagStream&   operator<< (OTagStream&, const TransformationData&);
ITagStream&   operator>> (ITagStream&, TransformationData&);

std::ostream& operator<< (ostream&,    const list<TransformationData>&);
OTagStream&   operator<< (OTagStream&, const list<TransformationData>&);
ITagStream&   operator>> (ITagStream&, list<TransformationData>&);


class PerspectiveTransformation
{
public:

	PerspectiveTransformation(){}
	~PerspectiveTransformation(){}


	
	// transform grayscale image with transformation
	void processImg (TransformationData &trans, Image<uint8>& srcDest, bool linearInterpolation=true)
	{
		Image<uint8> tmp = srcDest;
		processImg (trans, tmp, srcDest, linearInterpolation);
	} 

	// transform grayscale image with transformation
	void processImg (TransformationData &trans, const Image<uint8>& src, Image<uint8>& dest, bool linearInterpolation=true)
	{
		if (src.getSize()[DIM_BAND] != 1) {
			cerr << "error in: PerspectiveTransformation: processImg()"	<< endl 
				 << "only grayscale images can be transformed!" << endl << endl;
			exit(1);
		} // if

		ImageSize size = dest.getSize();
		if ((size[DIM_X]==0) || (size[DIM_Y] == 0)) {
			size = src.getSize();
			dest.resize(size);
		} // if

		if (linearInterpolation) {
			TransformBitmap_LinInterpol (src, dest, size, trans);
		} else {
			TransformBitmap_NonLinInterpol (src, dest, size, trans);
		}
	} 


	// transform color (RGB) image with transformation
	void processColorImg (TransformationData &trans, Image<uint8>& srcDest, bool linearInterpolation=true)
	{
		Image<uint8> tmp = srcDest;
		processColorImg (trans, tmp, srcDest, linearInterpolation);
	} 

	// transform color (RGB) image with transformation
	void processColorImg (TransformationData &trans, const Image<uint8>& src, Image<uint8>& dest1, bool linearInterpolation=true)
	{
		if (src.getSize()[DIM_BAND] != 3) {
			cerr << "error in: PerspectiveTransformation: processImg()"	<< endl 
				 << "only color (RGB) images can be transformed!" << endl << endl;
			exit(1);
		} // if


		// convert RGB to YUV image --> this creates artefacts
		//Image<uint8> yuv;
		//ConvertColorSpace::process(CODEC_RGB, CODEC_YUV444, src, yuv);
		
		// initialize destImg
		ImageSize destSize = dest1.getSize();
		if ((destSize[DIM_X]==0) || (destSize[DIM_Y] == 0)) {
			destSize = src.getSize();
			dest1.resize(destSize);
		} // if
		destSize[DIM_BAND]=1;
		Image<uint8> destImg(destSize);


		// imitialize img
		ImageSize size = src.getSize();
		size[DIM_BAND] = 1;
		Image<uint8> img (size);


		// process Y plane
		img = src.getImageBand (0); 
		processImg (trans, img, destImg, linearInterpolation);
		dest1.setImageBand (destImg,0);


		// process U plane
		img = src.getImageBand (1); 
		processImg (trans, img, destImg, linearInterpolation);
		dest1.setImageBand (destImg,1);

		// process V plane
		img = src.getImageBand (2); 
		processImg (trans, img, destImg, linearInterpolation);
		dest1.setImageBand (destImg,2);

		// store in dest --> removed
		//ConvertColorSpace::process(CODEC_YUV444, CODEC_RGB, yuv, dest1);
	} 


	// transform image from src to dest with transformation
	void TransformBitmap_LinInterpol (
		const Image<uint8>& src, 
		Image<uint8>& dest,
		ImageSize &destSize,
		TransformationData &trans, 
		uint8 background=255,                                // color-value at undefined positions
		int32 dest_left =0,int32 dest_top=0,               // borders of destination image
		int32 dest_right=999999,int32 dest_bottom=999999); // borders of destination image


	// transform image from src to dest with transformation but no interpolation
	void TransformBitmap_NonLinInterpol (
		const Image<uint8>& src, 
		Image<uint8>& dest,
		ImageSize &destSize,
		TransformationData &trans, 
		uint8 background=255,                                // color-value at undefined positions
		int32 dest_left =0,int32 dest_top=0,               // borders of destination image
		int32 dest_right=999999,int32 dest_bottom=999999); // borders of destination image
														   
	

	// set parameters to original parameters, if they differ only slightly from original values
	void normalizeParameters(TransformationData &trans)
	{
		const ieee64 aDelta = 8e-6;
		const ieee64 pDelta = 1.9e-8;

		if (fabs(trans.a00-1.0) < aDelta) trans.a00=1.0;
		if (fabs(trans.a11-1.0) < aDelta) trans.a11=1.0;
		if (fabs(trans.a10)     < aDelta) trans.a10=0.0;
		if (fabs(trans.a01)     < aDelta) trans.a01=0.0;

		if (fabs(trans.px) < pDelta) trans.px=0.0;
		if (fabs(trans.py) < pDelta) trans.py=0.0;
	}


public:
};


#endif
