/*
   CameraTransformationPrecise.cpp     get precise camera transformation parameters
    Copyright (C) 2004 Stephan Kopf

	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// MoCA includes
#include <CameraTransformationPrecise.h>



void 
CameraTransformationPrecise::process (const Image<uint8> &refImg, const Image<uint8> &img, const Image<uint8> &mask,
									  const TransformationData &srcTrans,  
									  TransformationData &destTrans, ieee64 &minErr)
{
	// initialize variables
	ref_w = refImg.getSize()[DIM_X];
	ref_h = refImg.getSize()[DIM_Y];

	img_w = img.getSize()[DIM_X];
	img_h = img.getSize()[DIM_Y];

	this->refImg = refImg;
	this->img = img;
	this->mask = mask; // this is the mask fitting to the img

	lambda=0.001;
	minerror_valid=false;
	minerror = 0;



	destTrans = srcTrans;
	
	int cnt=0;
	isError=false;
	while (cnt < 50 && !AtOptimum() && (!isError))
    {
		//cout << "cnt: " << cnt << " lambda: " << lambda << " minerror: " << minerror << endl;
		try {
			transformationStep (destTrans);
		} catch(...) {
			isError = true;
			break;
		} // catch
		cnt++;
		
		if (cnt>50) {
			isError=true;
			break;
		} // if
    }
	if (isError) {
		destTrans = srcTrans;
	} // if

	minErr = minerror;
	return;
} // preocess


ieee64
CameraTransformationPrecise::calcError(const double *s)
{
	TransformationData t = valueToTrans(s);
	
	uint32 cnt=0;
	double err=0.0;
	for (int y=0;y<img_h;y++) {
		for (int x=0;x<img_w;x++) {

			if (mask(x,y)!=255) { // not object
				double de = 1.0/(x*t.px+y*t.py+1);
				double ix = (x*t.a00+y*t.a01+t.tx)*de;
				double iy = (x*t.a10+y*t.a11+t.ty)*de;
				double refpix;
				
				if ((ix>=0) && (ix<ref_w-1) && (iy>=0) && (iy<ref_h-1) && ((refpix=refGetPixel(iy,ix))!=255)) {
					err += (refpix-img(x,y)) * (refpix-img(x,y));
					cnt++;
				} // if
				
			} // if
		} // for
	} // for

	if (cnt <= 0) cnt=10000000;
	return err/cnt;
} // calcError



ieee64 
CameraTransformationPrecise::refGetPixel (ieee64 y,ieee64 x)
{
	int32 ix = (int32)x;
	int32 iy = (int32)y;
	double x0 = ix;
	double y0 = iy;
	double wx = x-x0;
	double wy = y-y0;
	
	if (ix<0 || ix>=ref_w-1 || iy<0 || iy>=ref_h-1) {
		return (ix+iy)%5*63;
	} // if
	
	return ((1-wy) * ((1-wx)*refImg(ix,iy)   + wx*refImg(ix+1,iy)) +
		       wy  * ((1-wx)*refImg(ix,iy+1) + wx*refImg(ix+1,iy+1)));
//	return ((1-wy)*((1-wx)*p[iy  ][ix] + wx*p[iy  ][ix+1]) +
   //		   wy *((1-wx)*p[iy+1][ix] + wx*p[iy+1][ix+1]));
}


void 
CameraTransformationPrecise::calcAll (const ieee64* state)
{
	TransformationData t = valueToTrans(state);
	PerspectiveTransformation perspTrans;
	
	uint32 cnt=0;
	
	ieee64 h[8][8];
	
	for (int i=0;i<8;i++) gradient[i]=0.0;

	for (int i=0;i<8;i++) {
		for (int j=0;j<8;j++) {
			h[i][j]=0.0;
		} // for
	} // for
		
	ieee64 totalerr=0.0;
	ieee64 e[8];
		
//	for (int y=0;y<img_h-1;y++) {
//		for (int x=0;x<img_w-1;x++) {
	for (int y=0;y<img_h;y++) {
		for (int x=0;x<img_w;x++) {
			if (mask(x,y) != 255) {
				TransformationPoint<ieee64> src,dest;
				src.x = x;
				src.y = y;
				TransformationData::processPoint(t, src, dest);
				double xi = dest.x;
				double yi = dest.y;
				
				if (xi<1 || xi>=ref_w-2 || yi<1 || yi>=ref_h-2) {
					continue;
				} // if
				
				double D   = 1.0/(t.px*x+t.py*y+1);
				
				double err,I1x,I1y;
				refGetPixelGrads (yi,xi, err,I1x,I1y);
				err -= img(x,y);
				
				if (fabs(err)> 20.0) D=0.0;
				
				totalerr += err*err;
				
				e[0] = x*D*I1x;
				e[1] = y*D*I1x;
				e[2] = 1*D*I1x;
				e[3] = x*D*I1y;
				e[4] = y*D*I1y;
				e[5] = 1*D*I1y;
				e[6] = -x*D*(xi*I1x+yi*I1y);
				e[7] = -y*D*(xi*I1x+yi*I1y);
				
				for (int i=0;i<8;i++) {
					gradient[i] += err*e[i];
				} // for
				
				for (int i=0;i<8;i++) {
					for (int j=i;j<8;j++) {
						h[i][j] += e[i]*e[j];
					} // for
				} // for
				
				cnt++;
			} // if
		} // for
	} // for
			
	if (cnt<=100) {
		isError=true;
		return;
	} // if

	for (int i=0;i<8;i++) {
		for (int j=i;j<8;j++) {
			hesse[i][j] = h[i][j];
			hesse[j][i] = h[i][j];
		} // for
		
		for (int i=0;i<8;i++) {
			gradient[i] *= 2; // /= cnt;
		} // for
	} // for
	return;
}

void
CameraTransformationPrecise::refGetPixelGrads (ieee64 y,ieee64 x, ieee64& err,ieee64& I1x, ieee64& I1y)
{

	int32 ix = (int32)x;
	int32 iy = (int32)y;
	ieee64 x0 = ix;
	ieee64 y0 = iy;
	ieee64 wx = x-x0;
	ieee64 wy = y-y0;
	
	err = ((1-wy)*((1-wx)*refImg(ix,iy)   + wx*refImg(ix+1,iy)) +
		      wy *((1-wx)*refImg(ix,iy+1) + wx*refImg(ix+1,iy+1)));
//	err = ((1-wy)*((1-wx)*p[iy  ][ix] + wx*p[iy  ][ix+1]) +
//		      wy *((1-wx)*p[iy+1][ix] + wx*p[iy+1][ix+1]));
	
	I1x = ((1-wy)*((1-wx)*(refImg(ix+1,iy)  -(int32)refImg(ix-1,iy))   + wx*(refImg(ix+2,iy)  -(int32)refImg(ix,iy))) +
		      wy *((1-wx)*(refImg(ix+1,iy+1)-(int32)refImg(ix-1,iy+1)) + wx*(refImg(ix+2,iy+1)-(int32)refImg(ix,iy+1))))/2;
//	I1x = ((1-wy)*((1-wx)*(p[iy  ][ix+1]-(int)p[iy  ][ix-1]) + wx*(p[iy  ][ix+2]-(int)p[iy  ][ix])) +
//		      wy *((1-wx)*(p[iy+1][ix+1]-(int)p[iy+1][ix-1]) + wx*(p[iy+1][ix+2]-(int)p[iy+1][ix])))/2;
	
	I1y = ((1-wy)*((1-wx)*(refImg(ix,iy+1)-(int32)refImg(ix,iy-1)) + wx*(refImg(ix+1,iy+1)-(int32)refImg(ix+1,iy-1))) +
		      wy *((1-wx)*(refImg(ix,iy+2)-(int32)refImg(ix,iy))   + wx*(refImg(ix+1,iy+2)-(int32)refImg(ix+1,iy  ))))/2;
//	I1y = ((1-wy)*((1-wx)*(p[iy+1][ix]-(int)p[iy-1][ix]) + wx*(p[iy+1][ix+1]-(int)p[iy-1][ix+1])) +
//		      wy *((1-wx)*(p[iy+2][ix]-(int)p[iy  ][ix]) + wx*(p[iy+2][ix+1]-(int)p[iy  ][ix+1])))/2;
	
	return;
}


void 
CameraTransformationPrecise::transformationStep (TransformationData &trans)
{
	ieee64 state[8];
	state[0] = trans.a00;
	state[1] = trans.a01;
	state[2] = trans.tx;
	state[3] = trans.a10;
	state[4] = trans.a11;
	state[5] = trans.ty;
	state[6] = trans.px;
	state[7] = trans.py;
	
	if (!minerror_valid)
    {
		minerror = calcError(state);
		minerror_valid=true;
    }


	calcAll(state);
	if (isError) {
		return;
	} // if
	
	gsl_vector* rhs      = gsl_vector_alloc(8);
	gsl_vector* solution = gsl_vector_alloc(8);
	gsl_matrix* matrix   = gsl_matrix_alloc(8,8);
	
	
	for (int i=0;i<8;i++) {
		for (int j=0;j<8;j++) {
			double val = hesse[i][j];
			if (i==j) val += lambda*val;
			
			gsl_matrix_set(matrix,i,j, val);
		} // for
	} // for
		
	for (int j=0;j<8;j++) {
		gsl_vector_set(rhs,j,-gradient[j]);
	} // for
		
		
	if (!affine) {
		delparam(0,matrix,rhs);
		delparam(1,matrix,rhs);
		delparam(3,matrix,rhs);
		delparam(4,matrix,rhs);
	} // if
		
	if (!persp) {
		delparam(6,matrix,rhs);
		delparam(7,matrix,rhs);
	} // if
		
		
	gsl_linalg_HH_solve(matrix,rhs,solution);
	
	ieee64 newstate[8];
	for (int i=0;i<8;i++) {
		newstate[i] = state[i];
	} // for

	ieee64* s = newstate;  
	for (int i=0;i<8;i++) {
		s[i] += gsl_vector_get(solution,i);
	} // for
	
	gsl_vector_free(rhs);
	gsl_vector_free(solution);
	gsl_matrix_free(matrix);
		
	ieee64 newval = calcError(newstate);
	if (newval<minerror) {

		minerror = newval;
		for (int i=0;i<8;i++) {
			state[i] = newstate[i];
		} // for
		if (lambda > 1e-9) {
			lambda /= 10;
		} // if

	} else {
		if (lambda<1e9) {
			lambda *= 10;
		} // if
	} // else
		
//		for (int i=0;i<8;i++)
//			cout << state[i] << " ";
//		cout << "lambda: " << lambda << endl;
		
	trans = valueToTrans(state);
	return;
}

