/*
    CssMaximaFeature.cpp     extract maxima in CSS images
    Copyright (C) 2004 Gerald Kuehne

	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <CssMaximaFeature.h>

OTagStream& operator<< (OTagStream& os, const CssMaximumItem& cm)
{
    TagFormat::writeStartTag(os, "CssMaximum");
	TagFormat::writeStartTag(os, "position");   os << cm.position;   TagFormat::writeEndTag(os, "position");
	TagFormat::writeStartTag(os, "height");     os << cm.height;     TagFormat::writeEndTag(os, "height");
	TagFormat::writeStartTag(os, "width");      os << cm.width;      TagFormat::writeEndTag(os, "width");
	TagFormat::writeStartTag(os, "curveStart"); os << cm.curveStart; TagFormat::writeEndTag(os, "curveStart");
	TagFormat::writeStartTag(os, "curveEnd");   os << cm.curveEnd;   TagFormat::writeEndTag(os, "curveEnd");
	TagFormat::writeStartTag(os, "isConcave");  os << cm.isConcave;  TagFormat::writeEndTag(os, "isConcave");
    TagFormat::writeEndTag (os, "CssMaximum");

    return os;
}

ITagStream& operator>> (ITagStream& is, CssMaximumItem& cm)
{
    TagFormat::readStartTag(is, "CssMaximum");
	TagFormat::readStartTag(is, "position");   is >> cm.position;   TagFormat::readEndTag(is, "position");
	TagFormat::readStartTag(is, "height");     is >> cm.height;     TagFormat::readEndTag(is, "height");
	TagFormat::readStartTag(is, "width");      is >> cm.width;      TagFormat::readEndTag(is, "width");
	TagFormat::readStartTag(is, "curveStart"); is >> cm.curveStart; TagFormat::readEndTag(is, "curveStart");
	TagFormat::readStartTag(is, "curveEnd");   is >> cm.curveEnd;   TagFormat::readEndTag(is, "curveEnd");
	TagFormat::readStartTag(is, "isConcave");  is >> cm.isConcave;  TagFormat::readEndTag(is, "isConcave");
    TagFormat::readEndTag (is, "CssMaximum");

    return is;
}

OTagStream& operator<< (OTagStream& os, const CssMaximumData& src)
{
	TagFormat::writeNodeValue(os, "xyRation",src.xyRatio);
	int n=0;
    for (vector<CssMaximumItem>::const_iterator it = src.data.begin(); it != src.data.end(); it++) {
		if (n>=DEFAULT_CSS_PEAK_NUMBER) {
			break;
		} // if
        os << (*it);
		n++;
    } // for

    return os;
}

ITagStream& operator>> (ITagStream& is, CssMaximumData& dest)
{
    dest.data.clear();
    CssMaximumItem maximum;

    char c;

    bool failFlag = false;
	TagFormat::readNodeValue(is, "xyRation",dest.xyRatio);


	int32 n=0;
    while (is >> c) {
        is.putback(c);
		if (n>=DEFAULT_CSS_PEAK_NUMBER) {
			break;
		} // if
        if (is >> maximum) {
            dest.data.push_back (maximum);
        } else {
            failFlag = true;
        }
		n++;
    } // while

    if (!failFlag) {
        is.clear(is.rdstate() & ~ios::failbit);
    }

    return is;
}



void
CssMaximaFeature::
process (const Image<uint8>& src, CssMaximumData& dest, vector<ImagePosition> &contourPixel)
{
	// get parameters of contour
	int32 pxMin=10000000;
	int32 pxMax=-1;
	int32 pyMin=10000000;
	int32 pyMax=-1;
	int32 i;
	vector<ImagePosition>::iterator it = contourPixel.begin();
	for (i=0; i< (int32)contourPixel.size(); i++) {
		int32 px = (*it)[DIM_X];
		int32 py = (*it)[DIM_Y];
		if(pxMin > px) pxMin = px;
		if(pxMax < px) pxMax = px;
		if(pyMin > py) pyMin = py;
		if(pyMax < py) pyMax = py;
		it++;
	} // for
	dest.xyRatio = (pyMax-pyMin);
	if (dest.xyRatio > 0.0) {
		dest.xyRatio = (pxMax-pxMin) / dest.xyRatio;
	} else {
		dest.xyRatio = 1.0;
	}


    processBottomWidth (src, dest);
}



/*!  Extracts maxima (value, position, curve start and curve end) from
  the curves of a CSS image. Curve start and end are given for the maxwidth
  position.

  \param src  CSS image
  \param dest vector for maxima structs
*/

void
CssMaximaFeature::
processMaxWidth (const Image<uint8>& src, CssMaximumData& dest)
{
    assert(src.getBorderMode() == BORDER_PERIODIC);

    /* Step 1: scan the last line of the CSS image and store the
       start and end points of the curves.  It might be necessary to
       swap start and end point since CSS curves can cross the image
       border (see Step 3). */

    CssMaximumItem cssCurve[MAX_NUMBER_CSS_MAXIMA];

    // initialize, -1 indicates that the curve has not been processed yet
	indexType i;
    for (i = 0; i < MAX_NUMBER_CSS_MAXIMA; i++) {
        cssCurve[i].curveStart = -1;
        cssCurve[i].curveEnd   = -1;
    }

    // scan the last line of the CSS image
    ImageSize size = src.getSize();

    for (i = 0; i < size[DIM_X]; i++) {
        // get the label of the pixel
        uint8 label = src(i, size[DIM_Y] - 1);
        if (label == BACKGROUND)
            continue;
        // we have a non-background pixel, i.e. a start or end point of a CSS curve
        if (cssCurve[label].curveStart == -1) {
            // its the start pixel of the curve (not true when CSS curve crosses
            // the image border)
            cssCurve[label].curveStart = i;

        } else {
            // its the end pixel (not true when CSS curve crosses the image border)
            cssCurve[label].curveEnd = i;
        }
    }

    /* Step 2: find the maxima */

    findMaxima(src, cssCurve);

    /* Step 3: recalculate start and end points to get maximum distance */

    // store, which curves have already been processed
    bool curveProcessed[MAX_NUMBER_CSS_MAXIMA];
    for (i = 0; i < MAX_NUMBER_CSS_MAXIMA; i++) {
        curveProcessed[i] = false;
    }

    // scan the CSS image from top left to bottom right
    for (indexType y = 0; y < size[DIM_Y]; y++) {
        for (indexType x = 0; x < size[DIM_X]; x++) {
            uint8 label = src(x,y);
            // continue, if pixel is a background pixel or
            // if it belongs to an already processed curve
            if (label == BACKGROUND || curveProcessed[label])
                continue;

            // we found a curve , do not process it a second time
            // (there might be curves that cross the boundary multiple
            // times)
            curveProcessed[label] = true;

            // find start point and end point with largest distance of the curve
            traceCurveMaxWidth (src, cssCurve[label]);
        }

        // calculate width of the curve
        for (indexType i = 0; i < MAX_NUMBER_CSS_MAXIMA; i++) {
            // exists a curve with label i?
            if (cssCurve[i].curveStart != -1) {
                cssCurve[i].width = (cssCurve[i].curveEnd + size[DIM_X] -
                                     cssCurve[i].curveStart) % size[DIM_X];
            }
        }
    }

    // create the CSS maxima vector
    dest.data.clear();

    for (i = 0; i < MAX_NUMBER_CSS_MAXIMA; i++) {
        // exists a curve with label i?
        if (cssCurve[i].curveStart != -1) {
            dest.data.push_back(cssCurve[i]);
        }
    }
}


/*!  Extracts maxima (value, position, curve start and curve end) from
  the curves of a CSS image. Curve start and end are given for the bottom
  position.

  \param src  CSS image
  \param dest vector for maxima structs
*/

void
CssMaximaFeature::
processBottomWidth (const Image<uint8>& src, CssMaximumData& dest)
{
    assert(src.getBorderMode() == BORDER_PERIODIC);


    /* Step 1: scan the last line of the CSS image and store the
       start and end points of the curves.  It might be necessary to
       swap start and end point since CSS curves can cross the image
       border (see Step 3). */

    CssMaximumItem cssCurve[MAX_NUMBER_CSS_MAXIMA];

    // initialize, -1 indicates that the curve has not been processed yet
	indexType i;
    for (i = 0; i < MAX_NUMBER_CSS_MAXIMA; i++) {
        cssCurve[i].curveStart = -1;
        cssCurve[i].curveEnd   = -1;
    }

    // scan the last line of the CSS image
    ImageSize size = src.getSize();

    for (i = 0; i < size[DIM_X]; i++) {
        // get the label of the pixel
        uint8 label = src(i, size[DIM_Y] - 1);
        if (label == BACKGROUND)
            continue;
        // we have a non-background pixel, i.e. a start or end point of a CSS curve
        if (cssCurve[label].curveStart == -1) {
            // its the start pixel of the curve (not true when CSS curve crosses
            // the image border)
            cssCurve[label].curveStart = i;

        } else {
            // its the end pixel (not true when CSS curve crosses the image border)
            cssCurve[label].curveEnd = i;
        }
    }

    /* Step 2: find the maxima */

    findMaxima(src, cssCurve);


    /* Step 3: find the curves which cross or are at the image border
       and recalculate their start and end points */

    // store, which curves have already been processed
    bool curveProcessed[MAX_NUMBER_CSS_MAXIMA];
    for (i = 0; i < MAX_NUMBER_CSS_MAXIMA; i++) {
        curveProcessed[i] = false;
    }

    // check the image border for CSS curves
    for (indexType y = 0; y < size[DIM_Y]; y++) {
        uint8 label = src(0,y);
        // continue, if pixel is a background pixel or
        // if it belongs to an already processed curve
        if (label == BACKGROUND || curveProcessed[label])
            continue;

        // we found a curve at the border, do not process it a second time
        // (there might be curves that cross the boundary multiple times)
        curveProcessed[label] = true;

        // find start and end point of the curve
        traceCurve (src, cssCurve[label]);
    }

    // calculate width of the curve
    for (i = 0; i < MAX_NUMBER_CSS_MAXIMA; i++) {
        // exists a curve with label i?
        if (cssCurve[i].curveStart != -1) {
            cssCurve[i].width = (cssCurve[i].curveEnd + size[DIM_X] -
                                 cssCurve[i].curveStart) % size[DIM_X];
        }
    }

    // create the CSS maxima vector
    dest.data.clear();

    for (i = 0; i < MAX_NUMBER_CSS_MAXIMA; i++) {
        // exists a curve with label i?
        if (cssCurve[i].curveStart != -1) {
            dest.data.push_back(cssCurve[i]);
        }
    }
}


void
CssMaximaFeature::
findMaxima (const Image<uint8>& src, CssMaximumItem cssCurve[])
{
    ImageSize size = src.getSize();

    // store, which maxima have already been processed
    bool curveProcessed[MAX_NUMBER_CSS_MAXIMA];
    for (indexType i = 0; i < MAX_NUMBER_CSS_MAXIMA; i++) {
        curveProcessed[i] = false;
    }

    // scan the CSS image from top left to bottom right
    for (indexType y = 0; y < size[DIM_Y]; y++) {
        for (indexType x = 0; x < size[DIM_X]; x++) {
            // continue, if pixel is a background pixel or
            // if it belongs to an already processed curve
            uint8 label = src(x,y);

            if (label == BACKGROUND || curveProcessed[label])
                continue;

            // we found a new curve, do not process it a second time
            curveProcessed[label] = true;

            // set values; since we are scanning the image from top left to
            // bottom right we have found the top point of the curve
            // (more exact: the left top point, there might be more than one pixel
            // at this level)
            cssCurve[label].position = x;
            cssCurve[label].height   = size[DIM_Y] - y; // invert y position to get height
			if (label<128) {
				cssCurve[label].isConcave = true;       // value (concave part) < 128 
			} else {
				cssCurve[label].isConcave = false;      // value (concex part) >= 128 
			} // else
        } // for (indexType x ...)
    } // for (indexType y ...)
}


/*!  Traces a curve in the CSS image from top to bottom to find start
  and end point of the curve.

  \param src CSS image
  \param curve description of a CSS curve; must contain valid entries
  for position and height
*/

void
CssMaximaFeature::
traceCurve (const Image<uint8>& src, CssMaximumItem& curve)
{
    ImageSize size = src.getSize();

    /* find start position for the left part of the curve */

    // calculate left top point of the curve (must be provided in the parameter curve)
    ImagePosition startPosition (curve.position, size[DIM_Y] - curve.height);
    uint8 label = src(startPosition);

    ImagePosition startPositionLeftBranch(-1, -1);

    // search from left to right (in the range from -2 to +2 pixels)
    // in the row under the left top point for a pixel with the same
    // label
	indexType i;
    for (i = startPosition[DIM_X] - 2; i <= startPosition[DIM_X] + 2; i++) {
        if (src(i, startPosition[DIM_Y] + 1) == label) {
            startPositionLeftBranch = ImagePosition(i, startPosition[DIM_Y] + 1);
            break;
        }
    }

    // ensure that we found a start position
    assert(startPositionLeftBranch != ImagePosition(-1, -1));


    /* find start position for the right part of the curve */

    // calculate right top point of the curve
    // it's in the same row as the left top point
    while (src(startPosition[DIM_X] + 1, startPosition[DIM_Y]) == label) {
        startPosition[DIM_X]++;
    }

    ImagePosition startPositionRightBranch(-1, -1);

    // search from right to left in the row under the right top point for a
    // pixel with the same label
    for (i = startPosition[DIM_X] + 2; i >= startPosition[DIM_X] - 2; i--) {
        if (src(i, startPosition[DIM_Y] + 1) == label) {
            startPositionRightBranch = ImagePosition (i, startPosition[DIM_Y] + 1);
            break;
        }
    }

    // ensure that we found a start position
    assert(startPositionRightBranch != ImagePosition(-1, -1));


    /* trace down left and right part of the curve */

    ImagePosition posLeft  = startPositionLeftBranch;
    ImagePosition posRight = startPositionRightBranch;

    // repeat until last row is reached
    while (posLeft[DIM_Y] < size[DIM_Y] - 1) {
        // search in the neighbourhoud for the next curve pixel
        bool foundLeft = false;
		indexType i;
        for (i = posLeft[DIM_X] - 1; i <= posLeft[DIM_X] + 1; i++) {
            if (src(i, posLeft[DIM_Y] + 1) == label) {
                posLeft = ImagePosition(i, posLeft[DIM_Y] + 1);
                foundLeft = true;
                break;
            }
        }

        bool foundRight = false;
        for (i = posRight[DIM_X] + 1; i >= posRight[DIM_X] - 1; i--) {
            if (src(i, posRight[DIM_Y] + 1) == label) {
                posRight = ImagePosition(i, posRight[DIM_Y] + 1);
                foundRight = true;
                break;
            }
        }

        assert (foundLeft && foundRight);
    }

    // modify according to the image size
    curve.curveStart = (posLeft[DIM_X] + size[DIM_X]) % size[DIM_X];
    curve.curveEnd   = (posRight[DIM_X] + size[DIM_X]) % size[DIM_X];
}




/*!  Traces a curve in the CSS image from top to bottom to find start
  and end point with the maximum distance of the curve.

  \param src CSS image
  \param curve description of a CSS curve; must contain valid entries
  for position and height
*/

void
CssMaximaFeature::
traceCurveMaxWidth (const Image<uint8>& src, CssMaximumItem& curve)
{
    ImageSize size = src.getSize();

    /* find start position for the left part of the curve */

    // calculate left top point of the curve (must be provided in the parameter curve)
    ImagePosition startPosition (curve.position, size[DIM_Y] - curve.height);
    uint8 label = src(startPosition);

    ImagePosition startPositionLeftBranch(-1, -1);

    // search from left to right (in the range from -2 to +2 pixels)
    // in the row under the left top point for a pixel with the same
    // label
	indexType i;
    for (i = startPosition[DIM_X] - 2; i <= startPosition[DIM_X] + 2; i++) {
        if (src(i, startPosition[DIM_Y] + 1) == label) {
            startPositionLeftBranch = ImagePosition(i, startPosition[DIM_Y] + 1);
            break;
        }
    }

    // ensure that we found a start position
    assert(startPositionLeftBranch != ImagePosition(-1, -1));


    /* find start position for the right part of the curve */

    // calculate right top point of the curve
    // it's in the same row as the left top point
    while (src(startPosition[DIM_X] + 1, startPosition[DIM_Y]) == label) {
        startPosition[DIM_X]++;
    }

    ImagePosition startPositionRightBranch(-1, -1);

    // search from right to left in the row under the right top point for a
    // pixel with the same label
    for (i = startPosition[DIM_X] + 2; i >= startPosition[DIM_X] - 2; i--) {
        if (src(i, startPosition[DIM_Y] + 1) == label) {
            startPositionRightBranch = ImagePosition (i, startPosition[DIM_Y] + 1);
            break;
        }
    }

    // ensure that we found a start position
    assert(startPositionRightBranch != ImagePosition(-1, -1));


    /* trace down left and right part of the curve */

    ImagePosition posLeft  = startPositionLeftBranch;
    ImagePosition posRight = startPositionRightBranch;

    int16 maxCurveStart = 0;
    int16 maxCurveEnd   = 0;
    int16 maxWidth      = 0;

    // repeat until last row is reached
    while (posLeft[DIM_Y] < size[DIM_Y] - 1) {
        // search in the neighbourhoud for the next curve pixel
        bool foundLeft = false;
        for (i = posLeft[DIM_X] - 1; i <= posLeft[DIM_X] + 1; i++) {
            if (src(i, posLeft[DIM_Y] + 1) == label) {
                posLeft = ImagePosition(i, posLeft[DIM_Y] + 1);
                foundLeft = true;
                break;
            }
        }

        bool foundRight = false;
        for (i = posRight[DIM_X] + 1; i >= posRight[DIM_X] - 1; i--) {
            if (src(i, posRight[DIM_Y] + 1) == label) {
                posRight = ImagePosition(i, posRight[DIM_Y] + 1);
                foundRight = true;
                break;
            }
        }

        assert (foundLeft && foundRight);

        // calculate width for current curve points
        int16 curveStart = (posLeft[DIM_X] + size[DIM_X]) % size[DIM_X];
        int16 curveEnd   = (posRight[DIM_X] + size[DIM_X]) % size[DIM_X];
        int16 width      = (curveEnd + size[DIM_X] - curveStart) % size[DIM_X];

        if (width > maxWidth) {
            maxWidth = width;
            maxCurveStart = curveStart;
            maxCurveEnd  = curveEnd;
        }
    }

    // modify according to the image size
    curve.curveStart = maxCurveStart;
    curve.curveEnd   = maxCurveEnd;
}

