/*
    EdgeBasedContrastFeature.cpp     calculate the ECR-based contrast feature
    Copyright (C) 2004 Stephan Kopf

	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// MoCA includes
#include <EdgeBasedContrastFeature.h>


std::ostream& operator<< (ostream& os, const EbcData &data)
{
	os << "version        : " << data.version         << endl;
	os << "filename       : " << data.filename        << endl;
	os << "frames         : " << data.frames << endl;
	os << "threshold low  : " << (int32)data.thLow<< endl;
	os << "threshold high : " << (int32)data.thHigh<< endl;

	for (int32 i=0; i<data.data.size(); i++) {
		os << data.data[i] << endl;
	}

	return os;
} // operator<<

OTagStream& operator<< (OTagStream& os, const EbcData& data)
{
	
    TagFormat::writeStartTag (os, "edge-based-contrast");
	os << endl;
	
    TagFormat::writeNodeValue(os, "version", data.version, 1);
    TagFormat::writeNodeValue(os, "filename", data.filename, 1);
    TagFormat::writeNodeValue(os, "frames", data.frames, 1);
    TagFormat::writeNodeValue(os, "thLow", data.thLow, 1);
    TagFormat::writeNodeValue(os, "thHigh", data.thHigh, 1);

    TagFormat::writeStartTag(os, "data", 1);
	os << endl;
	for (int32 i=0; i<data.frames; i++) {
		os << data.data[i] << endl;
	}
    TagFormat::writeEndTag(os, "data", 1);
	os << endl;
	
    TagFormat::writeEndTag(os, "edge-based-contrast");
	os << endl;
	
    return os;
} // operator<<


ITagStream& operator>> (ITagStream& is, EbcData& data)
{
    bool failFlag = false;
    TagFormat::readStartTag(is, "edge-based-contrast");
	
    TagFormat::readNodeValue(is, "version", data.version);
    TagFormat::readNodeValue(is, "filename", data.filename);
    TagFormat::readNodeValue(is, "frames", data.frames);
    TagFormat::readNodeValue(is, "thLow", data.thLow);
    TagFormat::readNodeValue(is, "thHigh", data.thHigh);
	if (is.rdstate() & ios::failbit) {
		failFlag = true;
	} // if

	data.data.resize(data.frames);
    TagFormat::readStartTag(is, "data");
	ieee64 v;
	for (int32 i=0; i<data.frames; i++) {
		is >> v; data.data[i] = v;
	}
    TagFormat::readEndTag(is, "data");
    TagFormat::readEndTag(is, "edge-based-contrast");
	
	if (is.rdstate() & ios::failbit) {
		failFlag = true;
	} // if
    if (failFlag) {
		throw MocaException ("EdgeBasedContrast: Illegal index file!");
	} // if
	
    return is;
} // operator>> 




void 
EbcFeature::process(EbcData &data) 
{
	// read data
	ReadMocaData<EbcData> reader (pathFilename, EDGE_BASED_CONTRAST_FILENAME);
	bool dataValid = reader.process (EDGE_BASED_CONTRAST_VERSION, data); 
	
	if (dataValid) {
		if (data.thLow  != thLow) dataValid=false;
		if (data.thHigh != thHigh) dataValid=false;
	} // if

	if (!dataValid) {
		
		// create new index
		cerr << "\rEdgeBasedContrast : Creating index file ...         " << endl;
		readValues(data);
		
		// save index file
		WriteMocaData<EbcData> writer (pathFilename, EDGE_BASED_CONTRAST_FILENAME);
		writer.process(data);
		cerr << "\rEdgeBasedContrast : Creating index file ... done         " << endl;
	} // if

	return;
} // process


void 
EbcFeature::readValues (EbcData &data) 
{
	// read information
	VideoCodec codec = CODEC_Y;
	VideoDecoder reader (pathFilename, 3, -1, codec);
	VideoSize size = reader.getSize();


	// get path of filename, do not create new directory
	CreateMocaPath mocaPath (pathFilename, false);
	data.filename = mocaPath.file;
	
	data.frames   = size[DIM_Z];
	data.version  = EDGE_BASED_CONTRAST_VERSION;
	data.data.resize(data.frames);
	data.thLow = thLow;
	data.thHigh = thHigh;

	
	Image<uint8> img;
	for (int i=0; i<size[DIM_Z]; i++) {
		img = reader.getImage(i);
		cerr << "\rframe: " << i << "    " << ends;

		int32 vLow  = 0;
		int32 vHigh = 0;
		int32 edge;

		for (int32 y=0; y<size[DIM_Y]-1; y++)	 {
			for (int32 x=0; x<size[DIM_X]-1; x++)	 {
				edge = max (  abs(img(x,y)-img(x+1,y)),  abs(img(x,y)-img(x,y+1)) );
				if (edge>=thLow) {
					if (edge<thHigh) {
						vLow += edge;
					} else {
						vHigh += edge;
					} // else
				} // if
			} 
		} 

		// calculate ebc
		ieee64 ebc = 1.0 + ((vHigh - vLow - 1.0) / (vHigh + vLow + 1.0));
		data.data[i] = ebc;
	} 

	return;
} // init



