/*
   This file is part of the MoCA video library.
   Name: RegionFeature.h
   Copyright (C) 2004 Oliver Schuster

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: schuster@informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

/*! \file



   \author    Oliver Schuster

              Lehrstuhl fuer Praktische Informatik IV
              Universitaet Mannheim
              68131 Mannheim
              Germany
              email: schuster@informatik.uni-mannheim.de

   (c) Lehrstuhl fuer Praktische Informatik IV, details in COPYRIGHT.txt

   $Id: RegionFeature.h,v 1.2 2004/03/05 15:21:24 kopf Exp $

*/

#ifndef FLOOD_FILTER_H
#define FLOOD_FILTER_H

// C includes
#include <math.h>

// C++ includes
#include <stack>
#include <list>
#include <algorithm>
#include <exception>

// MoCA includes
#include <Image.h>
#include <ImageSize.h>
#include <MocaExceptions.h>
#include <Pixel.h>
#include <ForEachAlgos.h>
#include <GrayscaleFilter.h>

/*!

  \brief Determine regions in an image.

  This class provides functionality to find regions in an image based
  on similarity of neighbouring pixels. A few methods are provided to
  manipulate the image based on the region information.

  A future extension could be to implement the region growing
  algorithm for signals and videos. This might provide a
  characteristic to extract motion and objects.

  \todo works only on images so far

*/

struct RegionPixelPosition
{
	RegionPixelPosition (int32 x, int32 y)
		: x(x), y(y) {};
	int32 x,y;
};

class RegionType
{
public:
	// constructor adds first pixel of a region
	RegionType(int32 x, int32 y, int32 c0, int32 c1=0, int32 c2=0)
		: x(x), y(y), c0(c0), c1(c1), c2(c2), numPixels(1) {};

	void addPixel()
	{
	    numPixels++;
	    return;
	} // addPixel	

	int32 x,y;
	int32 c0,c1,c2;  // color plane 0..2
	int numPixels;
}; // RegionType

template<typename T>
class RegionFeature
{
public:
    //! the default constructor
    RegionFeature(T threshold=1, Neighbours n=N4);

    //! destructor
    ~RegionFeature();

    //! The actual filter
    void process(const Image<T> &src);
    void process(const Image<T> &src, Image<uint32> &regions);

    //! The actual filter, largestObjectInImage is binary image 
    //! of the largest object (not background)
    void process(const Image<T> &src, Image<uint8> &largestObjectInImage, bool &objectFound)
	{
        // calculate regions
		objectFound = false;
		Image<uint32> regions;
		process(src, regions);

        // initialize largest object
		ImageSize size = src.getSize();
		size[DIM_BAND]=1;
		largestObjectInImage.resize(size);
		largestObjectInImage.clear(255);

		int32 vecSize = regionVector.size();
		if (vecSize > 1) {
			locateObject (regions, largestObjectInImage);
			objectFound = true;
		} // if
	}
    
	void getRegionVector(vector<RegionType> &regionVector)
	{
		regionVector = this->regionVector;
		return;
	} // getRegionVector


    //! Fill all regions except one, which is specified
    void fillAllExcept(Image<T> &src, uint32 backReg, T fillValue);

    //! Average regions
    void fillAverageColor(Image<T> &src);

	//! Fill one region with another region
	// The variables region and regionVector are updated.
	void replaceRegion(Image<uint32> &region, vector<RegionType> &regionVector, uint32 oldRegion, uint32 newRegion);

	//! Fill one region with another region
	// Region is updated, regionVector not!
	void replaceRegionIndex(Image<uint32> &region, uint32 oldRegion, uint32 newRegion);


	// update the region vector (the region is set to average region color)
	void updateRegions(const Image<uint8> &img, const Image<uint32> &regions, vector<RegionType> &regionVector);
	
    //! Number of largest region
    unsigned int getLargestRegion();
    unsigned int getSecondLargestRegion();

    //! Number of largest region with specified color in first band
    unsigned int getLargestRegion(T val);


	T getBackgroundColor() 
	{
		return backgroundColor;
	} // getBackgroundColor

	T getForegroundColor() 
	{
		return foregroundColor;
	} // getForegroundColor

private:
    //! locate largest object (not background) and
	//! create binary image with object and update region map
    void locateObject (Image<uint32> &regions, Image<uint8> &objectImage);
    void getRelevantRegions (Image<uint32> &regions, uint32 &objectRegion, uint32 &backgroundRegion);
    
    void processColor(const Image<T> &src);
    void processGray(const Image<T> &src);


    //! calculate distance
    T distance(Pixel<T>&, Pixel<T>&);

    //! threshold for distance
    T threshold;

	//! neighbourhood
	Neighbours neighbours;

    //! which pixel belongs to which region
    Image<uint32> region;
    vector<RegionType> regionVector;

	T backgroundColor;
	T foregroundColor;

}; // class RegionFeature


#include <RegionFeature.cpt>

#endif
