/*
    SadFeature.cpp     sum of absolute differences
    Copyright (C) 2004 Dirk Farin

	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <SadFeature.h>



std::ostream& operator<< (ostream& os, const SadData &data)
{
	os << "version        : " << data.version          << endl;
	os << "filename       : " << data.filename         << endl;
	os << "sad size       : " << data.size             << endl;
	os << "minimum value  : " << data.minSadValue      << endl;
	os << "maximum value  : " << data.maxSadValue      << endl;
	os << "average value  : " << data.avgSadValue      << endl;
	os << "Moca path      : " << data.mocaPath         << endl;
	os << endl;

	return os;
} // operator<<

OTagStream& operator<< (OTagStream& os, const SadData& data)
{
    TagFormat::writeNodeValue(os, "version",     data.version);
    TagFormat::writeNodeValue(os, "filename",    data.filename);
    TagFormat::writeNodeValue(os, "size",     data.size);
    TagFormat::writeNodeValue(os, "minSadValue", data.minSadValue);
    TagFormat::writeNodeValue(os, "maxSadValue", data.maxSadValue);
    TagFormat::writeNodeValue(os, "avgSadValue", data.avgSadValue);
	os << endl;

	// write sad data
    TagFormat::writeStartTag (os, "data");
	TagFormat::writeNodeValue(os, data.sad);
    TagFormat::writeEndTag (os, "data");

	return os;
} // operator<<

ITagStream& operator>> (ITagStream& is, SadData& data)
{
    char c;
    bool failFlag = false;

	if (is >> c) {
		is.putback(c);

		TagFormat::readNodeValue(is, "version",     data.version);
		TagFormat::readNodeValue(is, "filename",    data.filename);
		int32 tmp = 0;
		TagFormat::readNodeValue(is, "size", tmp);
		data.size = tmp;

		TagFormat::readNodeValue(is, "minSadValue", data.minSadValue);
		TagFormat::readNodeValue(is, "maxSadValue", data.maxSadValue);
		TagFormat::readNodeValue(is, "avgSadValue", data.avgSadValue);
	} // if

	data.sad.clear();
    if (failFlag) {
		throw MocaException ("SadFeature: Illegal file format!");
	} // if

	// write sad data
	try {
	    TagFormat::readStartTag (is, "data");
		TagFormat::readNodeValue(is, data.sad);
		TagFormat::readEndTag (is, "data");
	} catch (...) {
			failFlag = true;
	} // try
	if (data.sad.size() <= 0) {
		failFlag = true;
	} // if

    if (failFlag) {
		string s = "SadFeature: Illegal data in index file!";
		throw MocaException (s);
	} // if

    return is;
} // operator>> 



const string SadFeature::DEFAULT_SAD_FEATURE_FILE_NAME = "sad.idx";


void 
SadFeature::process (Image<uint8> &img1, Image<uint8> &img2, ieee64 &sad)
{
	ImageSize size = img1.getSize();
	ieee64 factor = 1.0 / (255.0 * size[DIM_X] * size[DIM_Y] * size[DIM_BAND]);

	sad=0.0;
	for (int32 b=0; b<size[DIM_BAND]; b++) {
		for (int32 y=0; y<size[DIM_Y]; y++) {
			for (int32 x=0; x<size[DIM_X]; x++) {
				sad += ABS(((ieee64)img1(x,y,b)) - img2(x,y,b));
			}
		}
	}

	sad *= factor;
	return;
}

	


void
SadFeature::process (SadData& data) 
{

	// read data
	pathFilename = data.pathFilename;
	ReadMocaData<SadData> reader (pathFilename, SAD_FEATURE_FILENAME);
	bool dataValid = reader.process (SAD_FEATURE_VERSION, data); 

	if (!dataValid) {

		// create new index
		cerr << "SadFeature: Creating index file ..." << endl;
		readValues(data);


		// save index file
		WriteMocaData<SadData> writer (pathFilename, SAD_FEATURE_FILENAME);
		writer.process(data);
		cerr << "SadFeature: Creating index file ...done" << endl;
	} // if

	return;
}



// calculate SAD values
void
SadFeature::readValues (SadData& data) 
{

	// open video
	VideoDecoder reader (pathFilename);
	size = reader.getSize();

    // set data
	data.size = size[DIM_Z];
    data.sad.resize(data.size);

	data.avgSadValue = 0;
	data.maxSadValue = -1;
	data.minSadValue = 10e10;

	data.version  = SAD_FEATURE_VERSION;
	CreateMocaPath pathReader (pathFilename, false);
	data.filename = pathReader.file;

	for (uint32 i=0; i<data.size; i++) {
		data.sad[i]  = 0;
	} // for

    // process images
    Image<uint8> img0, img1;
	cerr << endl;
    for (uint32 i= 0; i < size[DIM_Z]-1; i++) {
		cerr << "\rframe: " << i << "            " << ends;

		ieee64 v = 0.0;
		img0 = reader.getImage(i);
		img1 = reader.getImage(i+1);
		process (img0, img1, v);
		data.sad[i] = v;

		if (data.minSadValue > data.sad[i]) {
			data.minSadValue = data.sad[i];
		} // if
		if (data.maxSadValue < data.sad[i]) {
			data.maxSadValue = data.sad[i];
		} // if
		data.avgSadValue += data.sad[i];

	} // for
	data.avgSadValue /= data.size;

	return;
}

