/*
   This file is part of the MoCA video library.
   Name: CannyFilter.h
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: kopf@informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

/*! \file



   \author    Stephan Kopf

              Lehrstuhl fuer Praktische Informatik IV
              Universitaet Mannheim
              68131 Mannheim
              Germany
              email: kopf@informatik.uni-mannheim.de

   (c) Lehrstuhl fuer Praktische Informatik IV, details in COPYRIGHT.txt

*/


#ifndef CANNY_FILTER_H
#define CANNY_FILTER_H

// C++ includes
#include <cmath>
#include <queue>
using namespace std;


// MoCA includes
#include <ImageFilterBase.h>
#include <Image.h>
#include <GaussFilter.h>
#include <ConvertFilter.h>
#include <DerivationFilter.h>
#include <GaussDerivationFilter.h>
#include <GrayscaleFilter.h>



template<class T>
class CannyFilter : public ImageFilterBase<T>
{
public:
	// @param noiseFraction specify noise level [0..1]
    //  @param suppress the percentage of noiseFraction edges pixels�, the value determines th th_low of the canny algorithm
    //  @param ori return the edge angle map (one band) if not 0
    //  @param vec returns the edge vector map (two bands) if not 0
    CannyFilter (int32 maskSize = 3)
    {
		init(maskSize, 0, 0, NULL, NULL);
	};
    CannyFilter (int32 maskSize, T th_low, T th_high)
    {
		init(maskSize, th_low, th_high, NULL, NULL);
	};
    CannyFilter (int32 maskSize, T th_low, T th_high,
                 Image<ieee64> *ori, Image<ieee64> *vec)

    {
		init(maskSize, th_low, th_high, ori, vec);
	};

    ~CannyFilter (){}

	void process (Image<T> &srcDest)
	{
	    Image<T> tmp = srcDest;
	    process (tmp, srcDest);
	    return;
	}
	void process (const Image<T> &src, Image<T> &dest);
	void process (const Image<T> &src, Image<T> &dest, bool processColor)
	{
		color=processColor;
		process (src, dest);
		return;
	}
	void process (Image<T> &srcDest, bool processColor)
	{
		color=processColor;
		process(srcDest);
	    return;
	}

private:
	// functions
	void init(int32 maskSize, T th_low, T th_high, Image<ieee64> *ori, Image<ieee64> *vec)
	{
		checkValue (maskSize % 2, "CannyFilter: Can only process odd mask sizes.");
		this->maskSize = maskSize;

        this->ori = ori;
        this->vec = vec;
		color=false;
		magMax = 0.0;

		T tmp=0;
		minV=GetType::getMin(tmp);
		maxV=GetType::getMax(tmp);

		thLow  = th_low;
		thHigh = th_high;
		calcTh=false;
		if ((thHigh==0) && (thLow==0)) calcTh=true;
		return;
	}
    void EstimateNoise (Image<ieee64>& mag);
    
    void NonMaximalSuppress (Image<ieee64>& xgrad, Image<ieee64>& ygrad, Image<ieee64>& mag,
                             Image<ieee64> *ori, Image<ieee64> *vec, Image<ieee64>& dest_float);
    void Hysterese (Image<ieee64>& src_float, Image<ieee64>& mag, Image<T>& dest);

    void CannyImage(const Image<T>& src, Image<T>& dest, 
                    Image<ieee64> *ori, Image<ieee64> *vec);
    void CannyXDImage(const Image<T>& src, Image<T>& dest, 
                      Image<ieee64> *ori, Image<ieee64> *vec);

	// variables
    int32 maskSize;    // required for gauss
    ieee64 noiseFraction;
    bool calcTh;
    Image<ieee64> *ori;
    Image<ieee64> *vec;    
	bool color;
	T minV;
	T maxV;
	ieee64 magMax;
	T thLow;
	T thHigh;
};

#include <CannyFilter.cpt>

#endif
