/*
   This file is part of the MoCA video library.
   Name: CropFilter.h
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: kopf@pi4.informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

/* -*- C++ -*-

   CropFilter.h  -- implementation of the gauss filter

   Author:    Stephan Kopf
              Lehrstuhl fuer Praktische Informatik IV
              Universitaet Mannheim, Germany
              email: kopf@pi4.informatik.uni-mannheim.de

*/



#ifndef CROP_FILTER_H
#define CROP_FILTER_H

// C includes
#include <math.h>

// MoCA includes
#include <ImageFilterBase.h>
#include <Image.h>
#include <ImagePosition.h>


/*
  Crops an image.

  ImagePosition p1: upper left corner (included in new image)

  ImagePosition p2: lower right corner (included in new image), image width=p2[DIM_Y]-p1[DIM_Y]+1
     or
  int width, int height: height and width of the cropped image

*/

template<class T>
class CropFilter: public ImageFilterBase<T>
{

public:
    CropFilter (ImagePosition &p1, ImagePosition &p2) {
		init (p1,p2);
    } // constructor
    CropFilter (const ImagePosition &p1, const ImagePosition &p2) {
		init (p1,p2);
    } // constructor

    CropFilter (ImagePosition &p1, int width, int height) {
		ImagePosition tmp(p1[DIM_X] + width-1, p1[DIM_Y] + height-1);	
		init (p1,tmp);
    } // constructor
    CropFilter (const ImagePosition &p1, int width, int height) {
		ImagePosition tmp(p1[DIM_X] + width-1, p1[DIM_Y] + height-1);	
		init (p1,tmp);
    } // constructor

    CropFilter (const ImageSize &imgSize, const int16 cropLeft, const int16 cropRight, const int16 cropTop, const int16 cropBottom) {
		ImagePosition tmp1(cropLeft,cropTop);	
		ImagePosition tmp2(imgSize[DIM_X] - cropRight  - 1, imgSize[DIM_Y] - cropBottom - 1);	
		init (tmp1,tmp2);
    } // constructor

    CropFilter (const int16 cropLeft, const int16 cropRight, const int16 cropTop, const int16 cropBottom) {
		p1[DIM_X] = cropLeft;
		p1[DIM_Y] = cropTop;

        this->cropRight  = cropRight;
        this->cropBottom = cropBottom;
		imgSizeKnown=false;
    } // constructor


    //! Apply crop filter to src, write out into dest
    void process (const Image<T>& src, Image<T>& dest);

    //! Apply crop filter to src, write into same image
    void process (Image<T>& srcDest)
	{
		Image<T> tmp=srcDest;
		process (tmp, srcDest);
		return;
	} // process

private:
    void init (const ImagePosition &p1, const ImagePosition &p2) {

		imgSizeKnown=true;
        this->p1 = p1;
        this->p2 = p2;

		this->cropRight  = 0;
		this->cropBottom = 0;
    } // init


    ImagePosition p1, p2;

	bool imgSizeKnown;
	int32 cropRight;
	int32 cropBottom;
}; // CropFilter

#include <CropFilter.cpt>

#endif

