/*
   This file is part of the MoCA video library.
   Name: OpticalFlowFilter.h
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: {kuehne,kopf}@pi4.informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


/* -*- C++ -*-

 OpticalFlowFilter.h  -- 
 
  Author:    Stephan Kopf
  Lehrstuhl fuer Praktische Informatik IV
  Universitaet Mannheim, Germany
  email: {kuehne,kopf}@pi4.informatik.uni-mannheim.de
  
   $Id: 
*/



#ifndef OPTICAL_FLOW_FILTER_H
#define OPTICAL_FLOW_FILTER_H

// C++ includes
#include <list>
#include <iostream>
using namespace std;

// MoCA includes
#include <MocaTypes.h>
#include <Image.h>
#include <DrawLine.h>
#include <EdgeFilter.h>
#include <GrayscaleFilter.h>
#include <BinarizeFilter.h>
#include <DrawDot.h>
#include <GetType.h>
//#include <OTagStream.h>
#include <ITagStream.h>





struct MacroBlockMotionVector
{
	MacroBlockMotionVector ()
	{
		diff = 0;
		sort = 0;
		colPosition = 0;
		rowPosition = 0;
		deleted = true;
	} // MacroBlockMotionVector 
	
    bool operator< (const MacroBlockMotionVector& mbmv) const
    {
        return sort < mbmv.sort;
    } // operator<

    bool operator== (const MacroBlockMotionVector& mbmv) const
    {
        return ((blockStart == mbmv.blockStart) && 
				(blockEnd == mbmv.blockEnd) && 
			    (motionVector == mbmv.motionVector) && 
				(diff == mbmv.diff) && 
				(sort == mbmv.sort) &&
				(colPosition == mbmv.colPosition) &&
				(rowPosition == mbmv.rowPosition) &&
				(deleted     == mbmv.deleted) );
    } // operator==

    MacroBlockMotionVector& operator= (const MacroBlockMotionVector& other) 
	{
		if (this != &other) {
			blockStart     = other.blockStart;
			blockEnd       = other.blockEnd;
			motionVector   = other.motionVector;
			diff           = other.diff;
			sort           = other.sort;
			colPosition    = other.colPosition;
			rowPosition    = other.rowPosition;
			deleted        = other.deleted;
		}
		return *this;
	} // operator =

	ImagePosition blockStart;				// upper left corner of macro block
	ImagePosition blockEnd; 				// lower right corner of macro block
	ImagePosition motionVector;				// motion vector
	ieee64 diff;							// difference
	ieee64 sort;							// sort operation
	int colPosition;						// position of column
	int rowPosition;						// position of row
	bool deleted;
};


struct OpticalFlowMotionVector
{
	OpticalFlowMotionVector()
	{
		mx = 0;
		my = 0;
	}
	OpticalFlowMotionVector(int mx, int my)
	{	
		this->mx = mx;
		this->my = my;
	}
	int mx;							// motion vector in x direction
	int my;							// motion vector in y direction
};


enum CameraOperationType {
	UNKNOWN_CAMERA_OPERATION = 0,
	PAN         			 = 1,
	TILT                     = 2,
	ZOOM_IN                  = 3,
	ZOOM_OUT                 = 4
};

ostream& operator<< (ostream&, const CameraOperationType&);
istream& operator>> (istream&, CameraOperationType&);

struct OpticalFlowCameraOperation {
	OpticalFlowCameraOperation() : operation(UNKNOWN_CAMERA_OPERATION), value(0), firstFrame(0), lastFrame(0) {}

    OpticalFlowCameraOperation& operator= (const OpticalFlowCameraOperation& other) 
	{
		if (this != &other) {
			operation  = other.operation;
			value      = other.value;
			firstFrame = other.firstFrame;
			lastFrame  = other.lastFrame;
		}
		return *this;
	} // operator =
	
	CameraOperationType operation;
	int value;
	int firstFrame;
	int lastFrame;
};

ostream& operator<< (ostream&, const OpticalFlowCameraOperation&);



const int DEFAULT_RESULT_SIZE=30;
struct OpticalFlowResult
{	
  OpticalFlowResult()
  {
    for (int i=0; i<DEFAULT_RESULT_SIZE; i++) {
      row    [i] = OpticalFlowMotionVector(0,0);
      column [i] = OpticalFlowMotionVector(0,0);
    } // for
    imageVector = OpticalFlowMotionVector(0,0);
    operation   = UNKNOWN_CAMERA_OPERATION;
    value       = 0;
  } // OpticalFlowResult
  OpticalFlowMotionVector row    [DEFAULT_RESULT_SIZE];
  OpticalFlowMotionVector column [DEFAULT_RESULT_SIZE];
  OpticalFlowMotionVector imageVector;
  CameraOperationType operation;
  int value;                       // size of operation, e. g. tilt
};


OTagStream& operator<< (OTagStream&, const OpticalFlowCameraOperation&);
ITagStream& operator>> (ITagStream&, OpticalFlowCameraOperation&);

OTagStream& operator<< (OTagStream&, const list<OpticalFlowCameraOperation>&);
ITagStream& operator>> (ITagStream&, list<OpticalFlowCameraOperation>&);



template<class T>
class OpticalFlowFilter 
{
public:
    OpticalFlowFilter () 
	{	
		macroBlockSize = 8;
		relevantBlocks = 50;
		noiseMotion = 20;
		imageSize = ImageSize(0,0);
	};
    OpticalFlowFilter (uint8 macroBlockSize) 
	{	
		this->macroBlockSize = macroBlockSize;
		relevantBlocks = 50;
		noiseMotion = 20;
		imageSize = ImageSize(0,0);
	};
    OpticalFlowFilter (uint8 macroBlockSize, int relevantBlocks) 
	{	
		this->macroBlockSize = macroBlockSize;
		this->relevantBlocks = relevantBlocks;
		noiseMotion = 20;
		imageSize = ImageSize(0,0);
	};
	~OpticalFlowFilter () {};
	
	//! Apply optical flow filter to images img1 and img2, write out into dest
	void process (const Image<T> &img1, const Image<T> &img2, Image<T> &dest);
    OpticalFlowCameraOperation getCameraOperation (int32 number)
    {
        return co[number];
    }
    
        
    
    
private:	
	void printLines (const Image<T> &img1, Image<T> &dest);

	
	void searchMotionVectors (const Image<T> &img1, const Image<T> &img2);
	void accumulateResults ();
	void calculateCameraOperation ();
	void display ();
	void createMacroBlocks (const Image<T> &img);
	ieee64 getValue (int dx, int dy, const Image<T> &img1, const Image<T> &img2, const MacroBlockMotionVector mv);
	

	int relevantBlocks; // percentage of all blocks
	uint8 macroBlockSize;
	int noiseMotion;    // percentage of motion vectors with noise
	list<MacroBlockMotionVector> mbmv;
	OpticalFlowResult ofr;
	ImageSize imageSize;
	OpticalFlowCameraOperation co [3];


	
}; // OpticalFlowFilter

#include <OpticalFlowFilter.cpt>

#endif

