/*
   This file is part of the MoCA video library.
   Name: OImageStream.h
   Copyright (C) 2004 Oliver Schuster

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: schuster@informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

/*! \file



   \author    Oliver Schuster

              Lehrstuhl fuer Praktische Informatik IV
              Universitaet Mannheim
              68131 Mannheim
              Germany
              email: schuster@informatik.uni-mannheim.de

   (c) Lehrstuhl fuer Praktische Informatik IV, details in COPYRIGHT.txt

   $Id: OImageStream.h,v 1.2 2004/03/05 15:21:24 kopf Exp $

*/

#ifndef O_IMAGE_STREAM_H
#define O_IMAGE_STREAM_H


// C++ includes
#include <algorithm>

// Moca includes
#include <MocaTypes.h>
#include <MocaExceptions.h>
#include <Image.h>
#include <OImageFormatStream.h>
#include <FormatParameter.h>
#include <ConvertFilter.h>
#include <FileHandling.h>


// Moca format stream includes
#include <OPgmFormatStream.h>
#include <OPpmFormatStream.h>
#include <OPngFormatStream.h>
#include <OJpegFormatStream.h>

/*!
  abstract base class for output image streams
*/
template<class T>
class OImageStream
{
public:
    //! the default constructor
    OImageStream() : os(NULL), ofs(NULL) {};

    //! constructor based on url provided
    OImageStream(const string &);

    //! open an output image stream
    void open(const string &);

    //! constructor based on url provided
    OImageStream(const string &, const FormatParameter &);

    //! open an output image stream
    void open(const string &, const FormatParameter &);

    //! the destructor
    ~OImageStream();

    //! close an output image stream
    void close() throw (MocaException);

    //! write an image to stream
    void write(const Image<T> &) throw (MocaException);

    //! write an image to stream
    OImageStream &operator<<(Image<T> &);

protected:

private:

    //! general output stream
    ostream *os;

    //! format specific output stream
    OImageFormatStream<T> *ofs;

    //! build "ofs" based on string
    void getFormatStream(const string &, const FormatParameter* = NULL);

}; // OImageStream



class WriteImage
{
public:
	// store uint8 images
	WriteImage (const Image<uint8> &img, const string &filename)
	{
		try {
			// open output filestream and write image
			OImageStream<uint8> ois(filename);
			ois.write(img);
		} catch (MocaException e) {
			cout << "WriteImage: " << e.what() << endl;
			exit (-1);
		} // try
	}


	// store ieee64 images and scale image by factor f
	WriteImage (const Image<ieee64> &img, const string &filename, ieee64 f=1.0)
	{
		try {
			// convert to uint8
			Image<uint8> intImg;
			ieee64 srcMax = 255/f;
			ConvertFilter<ieee64, uint8> convert(CONVERT_SCALE, 0, srcMax, 0, 255);
			convert.process(img, intImg);
			
			
			// open output filestream and write image
			OImageStream<uint8> ois(filename);
			ois.write(intImg);
		} catch (MocaException e) {
			cout << "WriteImage: " << e.what() << endl;
			exit (-1);
		} // try
	}


	// store int32 images and CROP image by factor f
	WriteImage (const Image<int32> &img, const string &filename)
	{
		try {
			// convert to uint8
			Image<uint8> intImg;
			ConvertFilter<int32, uint8> convert(CONVERT_CROP);
			convert.process(img, intImg);
			
			
			// open output filestream and write image
			OImageStream<uint8> ois(filename);
			ois.write(intImg);
		} catch (MocaException e) {
			cout << "WriteImage: " << e.what() << endl;
			exit (-1);
		} // try
	}

	// store int32 images and CROP image by factor f
	WriteImage (const Image<uint32> &img, const string &filename)
	{
		try {
			// convert to uint8
			Image<uint8> intImg;
			ConvertFilter<uint32, uint8> convert(CONVERT_CROP);
			convert.process(img, intImg);
			
			
			// open output filestream and write image
			OImageStream<uint8> ois(filename);
			ois.write(intImg);
		} catch (MocaException e) {
			cout << "WriteImage: " << e.what() << endl;
			exit (-1);
		} // try
	}


};

#include <OImageStream.cpt>

#endif
