/*
   This file is part of the MoCA video library.
   Name: VideoDecoder.h
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: kopf@pi4.informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

/* -*- C++ -*-

  VideoDecoder.h -- Interface for MPEG video
 
  Author: Stephan Kopf
  Lehrstuhl fuer Praktische Informatik IV
  Universitaet Mannheim, Germany
  email: kopf@pi4.informatik.uni-mannheim.de
  

*/

#ifndef VIDEO_DECODER_H
#define VIDEO_DECODER_H

// C++ includes
#include <string>
#include <vector>
#include <iostream>
using namespace std;

// MoCA includes
#include <MocaTypes.h>
#include <MocaSignal.h>
#include <VideoIndex.h>
#include <ConvertColorSpace.h>
#include <OpenProcess.h>
#include <IImageStream.h>
#include <GrayscaleFilter.h>



// this class opens the MPEG video and returns the requested frames
// the last decoded images are cached
const int DEFAULT_CACHED_VIDEOFRAMES=10;
class VideoDecoder
{
public:

	// constuctor defines filename of mpeg audio/video file
	VideoDecoder (const string &pathFile, VideoCodec cMode=CODEC_RGB);

	// constuctor defines filename of mpeg audio/video file, the number of frames 
	// to be cached, and the last frame to decode
	VideoDecoder (const string &pathFile, int cacheSize, int32 lastFrameToDecode, VideoCodec cMode=CODEC_RGB);

	// destructor
	~VideoDecoder ();
	
	VideoIndexData getIndex();
	Image<uint8>& getImage(int32 frame);

	// may require a lot of memory!
	//void getVideo (Video<uint8> &vid);


	// access to variables
	ImageSize getImageSize();
	VideoSize getSize();
    int32 getNumberOfFrames() 
    {
        return getSize()[DIM_Z];
    }
	ieee64 getFrameRate();
	int32 getBitRate();
	bool isRgb();
	bool isYuv();
	bool isY();



private:

	// initialization
	void init(const string &filename, int32 lastFrame);
	void initializeCache(int32 size);
	void initializeDecoder();

	// destruction
	void closeCache();
	void closeDecoder();

	// decode next image
	void nextImage         (Image<uint8> &img);
	void nextImageYuv      (Image<uint8> &img);
	void nextImageY        (Image<uint8> &img);

	
	int32 cacheHitNo (int32 frame);
	int32 imageInCache (int32 frame);


	// reader
	FILE* fh_in;
	string videodecodingExecutable;
	VideoCodec colorMode;

	// cache
	Image<uint8>* images;
	Signal<int32> cache;
	int cacheSize;
	int oldestCachePosition;
	int32 lastDecodedFrame;

	// file name and index
	string pathFilename;
	VideoIndexData index;

	// decoder
	int32 lastFrameToDecode;
	bool reachedEof;
	int32 skipBytes;
	int32 sx, sy; // original size of video (DIM_X, DIM_Y)
};



class ReadVideo
{
public:

	// read all images in AVI-/MPEG-video 
	//   example:  ReadVideo ("C:/temp/test.avi", vid);
	ReadVideo (const string &filename, Video<uint8> &vid)
	{
		process (filename, vid);
	}

	// read all images in directory
	//   example:  ReadVideo ("C:/temp/", "*.jpg", vid);
	ReadVideo (const string &path, const string &extension, Video<uint8> &vid, bool grayscale=false)
	{
		process (path, extension, vid, 0, 100000000, grayscale);
	}

	// read number of images in directory, starting with frameMin 
	//   example:  ReadVideo ("C:/temp/", "*.jpg", vid);
	ReadVideo (const string &path, const string &extension, Video<uint8> &vid, int32 frameMin, int32 number, bool grayscale=false)
	{
		process (path, extension, vid, frameMin, number, grayscale);
	}


private:
	void process (const string &filename, Video<uint8> &vid)
	{
		VideoIndexData data;
		try {
			// resize video
			VideoIndex index (filename,true);
			index.process(data);
			vid.resize(data.videoSize);
		} catch (MocaException e) {
			cerr << "ReadVideo: " << e.what() << endl;
			cerr << "(cannot allocate sufficient memory?)" << endl;
			exit (-1);
		} // try

		try {
			// read videos
			VideoDecoder reader (filename);
			for (int32 i=0; i<data.videoSize[DIM_Z]; i++) {
				vid.setImage (i, reader.getImage(i));
			} // for
		} catch (MocaException e) {
			cerr << "ReadVideo: " << e.what() << endl;
			exit (-1);
		} // try
		return;
	} // process


	void process (const string &path, const string &extension, Video<uint8> &vid, int32 frameMin, int32 number, bool grayscale)
	{

		// read all images in directory
		vector<FileInfo> files;
		const bool recurse = false;
		FileList fl (path, recurse);    
		fl.process(extension, files);
		if (files.size()<1) {
			cerr << "VideoDecoder::process: No files found in directory " << path << endl << endl;
			exit(1);
		} // if

		// define video size
		Image<uint8> img;
		{
		  string fn = files[0].name;
		  ReadImage (fn, img);
		}
		ImageSize srcSize = img.getSize();
		

		VideoSize size (srcSize[DIM_X], srcSize[DIM_Y], files.size(), srcSize[DIM_BAND]);
		if (size[DIM_Z]>number) size[DIM_Z]=number;
		if (grayscale) size[DIM_BAND]=1;


		try {
			// resize video
			vid.resize(size);
		} catch (MocaException e) {
			cerr << "ReadVideo: " << e.what() << endl;
			cerr << "(cannot allocate sufficient memory?)" << endl;
			exit (-1);
		} // try


		// read images 
		GrayscaleFilter<uint8> gray;
		try {
			// read videos
			for (int32 i=0; i<size[DIM_Z]; i++) {
				cerr << "\rread image: " << (i+frameMin) << "   " << ends;
				  {
				    string fn = files[i+frameMin].name;
				    ReadImage (fn, img);
				  }
				if (size[DIM_BAND] != srcSize[DIM_BAND]) gray.process(img);
				vid.setImage(i,img);
			} // for
		} catch (MocaException e) {
			cerr << "ReadVideo: " << e.what() << endl;
			exit (-1);
		} // try
		cerr << endl;
		return;
	} // process

};



#endif
