/*
   This file is part of the MoCA video library.
   Name: VideoIndex.h
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: kopf@pi4.informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

/* -*- C++ -*-

  MpegVideoDvdview.cpp -- Interface for MPEG1/MPEG2 video
  
	Author:    Stephan Kopf
	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim, Germany
	email: kopf@pi4.informatik.uni-mannheim.de
	
*/




#ifndef VIDEO_INDEX_H
#define VIDEO_INDEX_H

// C includes
#include <stdio.h>

// C++ includes
#include <string>
#include <fstream>
#include <iostream>
using namespace std;

// MoCA includes
#include <MocaTypes.h>
#include <Image.h>
#include <Video.h>
#include <FileList.h>
#include <WriteMocaData.h>
#include <OpenProcess.h>
#include <WaveHeader.h>
#include <ProcessReader.h>


// Windows includes
#ifdef COMPILER_MSVC
#  include <fcntl.h>
#  include <io.h>
#  include <process.h>
#else
#  ifdef CYGWIN
#    include <fcntl.h>
#    include <io.h>
#    include <process.h>
#  else
#    include <unistd.h>
#  endif
#endif

//*********************************************************************

const static string VIDEO_INDEX_VERSION  ="Version-0.1";
const static string VIDEO_INDEX_FILENAME ="videoindex.idx";
const string WINDOWS_VIDEODECODER_PROG_EXECUTABLE = "mocaplayer.exe";
const string LINUX_VIDEODECODER_PROG_EXECUTABLE   = "mocaplayer";



struct VideoIndexData
{
public:
	VideoIndexData () 
	{
		clear();
	}
    VideoIndexData& operator=(const VideoIndexData &other) {
		if (this != &other) {
			// video parameters
			filename    = other.filename;
			videoFormat = other.videoFormat;
			bitrate		= other.bitrate;
			width		= other.width;
			height		= other.height;
			fps			= other.fps;
			videoAspect = other.videoAspect;
			
			// audio parameters
			audio     = other.audio;
			audioData = other.audioData;
			
			// additional parameters
			version   = other.version;
			videoSize = other.videoSize;
		} // if
        return *this;
    } // operator=

	void clear()
	{
		// video parameters
		filename    = "";
		videoFormat = "";
		bitrate		= 0;
		width		= 0;
		height		= 0;
		fps			= 0.0;
		videoAspect = 0.0;

		// audio parameters
		audio = false;
		audioData.setHeader (0, 0, 0, 0);

		// additional parameters
		version             = VIDEO_INDEX_VERSION;
		videoSize[DIM_X]    = 0;
		videoSize[DIM_Y]    = 0;
		videoSize[DIM_Z]    = 0;
		videoSize[DIM_BAND] = 3;

		return;
	}

	ImageSize getImageSize() 
	{
		ImageSize size(videoSize[DIM_X], videoSize[DIM_Y], videoSize[DIM_BAND]);
		return size;
	} // getImageSize() 

	string getTimecode (int32 nFrames)
	{
		int hours   = (nFrames / ((int32)(fps * 60 * 60)) );
		int minutes = (nFrames / ((int32)(fps * 60     )) ) % 60;
		int seconds = (nFrames / ((int32)(fps          )) ) % 60;


		string s = "";
		char c;
		if (hours > 9) {
			c = (hours/10) + '0';			
			s += c;
		} else {
			s += "0";
		} // else
		c = (hours%10) + '0';			
		s += c;
		s += ":";

		if (minutes > 9) {
			c = (minutes/10) + '0';			
			s += c;
		} else {
			s += "0";
		} // else
		c = (minutes%10) + '0';			
		s += c;
		s += ":";
		
		if (seconds > 9) {
			c = (seconds/10) + '0';			
			s += c;
		} else {
			s += "0";
		} // else
		c = (seconds%10) + '0';			
		s += c;

		return s;
	}

	// video parameters
	//    ID_VIDEO_FORMAT=cvid
	//    ID_VIDEO_BITRATE=0
	//    ID_VIDEO_WIDTH=1088
	//    ID_VIDEO_HEIGHT=456
	//    ID_VIDEO_FPS=2.000
	//    ID_VIDEO_ASPECT=0.0000
	string videoFormat;
	int32  bitrate;
	int32  width;
	int32  height;
	ieee64 fps;
	ieee64 videoAspect;

	// audio parameters (the parameters of the audio codec are not stored)
	//    ID_AUDIO_CODEC=mp3
	//    ID_AUDIO_FORMAT=80
	//    ID_AUDIO_BITRATE=192000
	//    ID_AUDIO_RATE=44100
	//    ID_AUDIO_NCH=2
	bool audio;
	WaveHeader audioData;

	// additional parameters
	string filename;
	string version;
	VideoSize videoSize;   // width, height, number of frames (starting with frame 0)
};

std::ostream& operator<< (ostream&, const VideoIndexData&);
OTagStream& operator<< (OTagStream&, const VideoIndexData&);
ITagStream& operator>> (ITagStream&, VideoIndexData&);


// this class gets the parameters of an MPEG video (the audio and video track)
class VideoIndex
{

public:
	
	// constuctor setting filename
	VideoIndex (const string &pathFilename, bool silent=false, bool fastOpen=false, bool debugMsg=false);

	~VideoIndex ()
	{
		close();
	}
	
	// get the index structure
	void process(VideoIndexData &data);

	string getVideoDecoderCmd()
	{
		string cmd = LINUX_VIDEODECODER_PROG_EXECUTABLE;
#ifdef _WIN32
		cmd = WINDOWS_VIDEODECODER_PROG_EXECUTABLE;
#endif
		return cmd;
	} // getVideoDecoderCmd

private:
	
	void open(string &cmd);
	void close ();
	void readValues (VideoIndexData &data);
	bool readTextLines (vector<string> &textLines, string &cmd);
	void getVideoParameters (VideoIndexData &data);
	
	FILE* fh_in;
	string pathFilename;
	string videodecodingExecutable;
	bool silent;
	bool fastOpen;
	bool debugMsg;
};


  
#endif
