/*
   This file is part of the MoCA video library.
   Name: Image.h
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: kopf@pi4.informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

#ifndef IMAGE_H
#define IMAGE_H

// C++ includes
#include <list>

// MoCA includes
#include <SampleBlock.h>
#include <ImagePosition.h>
#include <ImageSize.h>
#include <Pixel.h>
#include <MocaExceptions.h>


template<class T>
class Image : public SampleBlock<T>
{
public:

    //template<class T> 
	//friend class SubImage;

    // Construction/Destruction
    Image (const ImageSize& size     = ImageSize (0, 0, 0),
           const ImageSize& tileSize = ImageSize (0, 0, 0),
           BorderMode borderMode     = BORDER_REPLICATE)
            : SampleBlock<T> (size, tileSize, borderMode) {}

    Image (const Image&);

    virtual ~Image ();

    Image& operator= (const Image&);

    T& operator() (indexType x, indexType y, indexType band);
    T& operator() (indexType x, indexType y);
    const T& operator() (indexType x, indexType y, indexType band) const;
    const T& operator() (indexType x, indexType y) const;

    //! return value at pos
    T& operator() (const ImagePosition &pos);

    //! return value at pos (const variant)
    const T& operator() (const ImagePosition &pos) const;

    // Interface
    void resize  (const ImageSize& size);
    void realloc (const ImageSize& size, const ImageSize& tileSize, BorderMode borderMode);

    ImageSize getSize () const;
    ImageSize getTileSize () const;

    const T* getSamplePtr (indexType x, indexType y, indexType band) const;
    const T* getSamplePtr (indexType x, indexType y) const;
    T* getSamplePtr (indexType x, indexType y, indexType band);
    T* getSamplePtr (indexType x, indexType y);

    T* getTilePtr   (indexType x, indexType y, indexType band);
    T* getTilePtr   (indexType x, indexType y);

    //! get the pixel implied in pixel
    Pixel<T> getPixel(Pixel<T> &pixel) const;

    //! get the pixel implied in pixel
    Pixel<T> getPixel(const indexType x, const indexType y) const;

    //! get the pixel at position pos
    Pixel<T> getPixel(const ImagePosition &pos) const;

    //! set the pixel
    void setPixel(const Pixel<T> &pixel);

    //! get all neighbours
    list<Pixel<T> >
    getNeighbours(const ImagePosition &pos, const int ng) const
        throw (MocaException);


	// get one image band
	Image getImageBand (int32 band) const
	{
		ImageSize s (this->size[DIM_X], this->size[DIM_Y], 1);
		Image<T> img (s);

		for (indexType y=0; y<s[DIM_Y]; y++) {
			const offsetType* off0 = &this->getSampleOffsetX (0);
			const offsetType* off1 = &img.getSampleOffsetX (0);

			T* ptr0 = (T*) SampleBlockBase::getSamplePtr (0, y, 0, band);
			T* ptr1 = (T*) img.getSamplePtr (0, y, 0);

			for (indexType x=0; x<s[DIM_X]; x++) {
				*ptr1 = *ptr0;

				ptr0 += *off0;  off0++;
				ptr1 += *off1;  off1++;
			} // for
		} // for

		return img;
	}

	// set one image band
	void setImageBand (const Image<uint8> &img, int32 band) 
	{
		ImageSize s = img.getSize();
		if (s[DIM_BAND]!=1) {
			cerr << "error in Image::setImageBand: an image with only one band is expected!" << endl;
			exit(1);
		} // if
		if ((s[DIM_X]!=this->size[DIM_X]) || (s[DIM_Y]!=this->size[DIM_Y])) {
			cerr << "error in Image::setImageBand: the images have a different size!" << endl;
			exit(1);
		} // if


		for (indexType y=0; y<s[DIM_Y]; y++) {
			const offsetType* off0 = &this->getSampleOffsetX (0);
			const offsetType* off1 = &img.getSampleOffsetX (0);

			T* ptr0 = (T*) SampleBlockBase::getSamplePtr (0, y, 0, band);
			T* ptr1 = (T*) img.getSamplePtr (0, y, 0);

			for (indexType x=0; x<s[DIM_X]; x++) {
				*ptr0 = *ptr1;

				ptr0 += *off0;  off0++;
				ptr1 += *off1;  off1++;
			} // for
		} // for

		return;
	}


}; // class Image



// handles a vector of images
template<class T>
class ImageVector
{
public: 
	ImageVector () {
		size = 0;
	}

	ImageVector (int32 n) {
		size = n;
		data.resize(size);
	}

    void resize (uint32 size) {
        this->size = size;
        data.resize(size);
        return;
 	}

    // get pointer to image
 	Image<uint8>* getImage (uint32 number) {
        if (number>=size) {
            number=size-1;
        } // if
        return &(data[number]);
 	}

 	void setImage (uint32 number, const Image<T> &img) {
        if (number>=size) {
            number=size-1;
        } // if

        // copy image
        data[number]=img;
        return;
    }

    // get size
 	int32 getSize () {
        return size;
 	}
    
    
private:
	int32 size;  // number of images
public:
	vector<Image<T> > data;
};


#include <Image.cpt>

#endif
