/*
   This file is part of the MoCA video library.
   Name: MocaSignal.h
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: kuehne@pi4.informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

/* -*- C++ -*-

   signal.h  -- data structure for one-dimensional, multiband signals

   Author:    Gerald Kuehne
              Lehrstuhl fuer Praktische Informatik IV
              Universitaet Mannheim, Germany
              email: kuehne@pi4.informatik.uni-mannheim.de

   $Id: Signal.h,v 1.2 2004/03/05 15:21:24 kopf Exp $
*/

#ifndef MOCA_SIGNAL_H
#define MOCA_SIGNAL_H

// C++ includes
#include <vector>


// MoCA includes
#include <SampleBlock.h>
#include <SignalPosition.h>
#include <SignalSize.h>


template<class T> class Signal : public SampleBlock<T>
{
public:
	//template<class S> 
	//friend class SubSignal;

    // Construction/Destruction
    Signal (const SignalSize& size     = SignalSize (0, 0),
            const SignalSize& tileSize = SignalSize (0, 0),
            BorderMode borderMode      = BORDER_REPLICATE);

    Signal (const Signal&);

    virtual ~Signal ();

    Signal& operator= (const Signal&);

    T& operator() (indexType x, indexType band);
    T& operator() (indexType x);
    const T& operator() (indexType x, indexType band) const;
    const T& operator() (indexType x) const;

    // Interface
    void resize  (const SignalSize& size);
    void realloc (const SignalSize& size, const SignalSize& tileSize, BorderMode borderMode);

    SignalSize getSize () const;
    SignalSize getTileSize () const;

    T* getSamplePtr (indexType x, indexType band);
    T* getSamplePtr (indexType x);
    T* getTilePtr   (indexType x, indexType band);
    T* getTilePtr   (indexType x);

	// TODO
    const T* getSamplePtr (indexType x, indexType band) const;
    const T* getSamplePtr (indexType x) const;


    //! get all neighbours
    list<SignalPosition>
    getNeighbours(const SignalPosition &pos, const int ng = 2) const
        throw (MocaException);

	//void minValue (SignalPosition &pos);
	//void maxValue (SignalPosition &pos);
};


template<class T>
inline
Signal<T>::
Signal (const SignalSize& size, const SignalSize& tileSize, BorderMode borderMode)
  : SampleBlock<T> (size, tileSize, borderMode)
{}

template<class T>
inline
Signal<T>::
Signal (const Signal<T>& other)
  : SampleBlock<T> (other)
{}

template<class T>
Signal<T>::
~Signal ()
{}

template<class T>
inline
Signal<T>&
Signal<T>::
operator= (const Signal<T>& other)
{
  SampleBlock<T>::operator= (other);
  return *this;
}


template<class T>
inline
T&
Signal<T>::
operator() (indexType x, indexType band)
{
  return (T&) *(this->sampleIndexX[x] + this->sampleIndexBand[band]);
}

template<class T>
inline
T&
Signal<T>::
operator() (indexType x)
{
  return (T&) *(this->sampleIndexX[x]);
}

template<class T>
inline
const T&
Signal<T>::
operator() (indexType x, indexType band) const
{
  return (T&) *(this->sampleIndexX[x] + this->sampleIndexBand[band]);
}

template<class T>
inline
const T&
Signal<T>::
operator() (indexType x) const
{
  return (T&) *(this->sampleIndexX[x]);
}

template<class T>
inline
void
Signal<T>::
realloc (const SignalSize& size, const SignalSize& tileSize, BorderMode borderMode)
{
  SampleBlock<T>::realloc (size, tileSize, borderMode);
}

template<class T>
inline
void
Signal<T>::
resize (const SignalSize& size)
{
  SampleBlock<T>::resize (size);
}


template<class T>
inline
SignalSize
Signal<T>::
getSize () const
{
  return SignalSize (this->size[DIM_X], this->size[DIM_BAND]);
}

template<class T>
inline
SignalSize
Signal<T>::
getTileSize () const
{
  return SignalSize (this->tileSize[DIM_X], this->tileSize[DIM_BAND]);
}

template<class T>
inline
T*
Signal<T>::
getSamplePtr (indexType x, indexType band)
{
  return (T*) (this->sampleIndexX[x] + this->sampleIndexBand[band]);
}

template<class T>
inline
T*
Signal<T>::
getSamplePtr (indexType x)
{
  return (T*) this->sampleIndexX[x];
}

template<class T>
inline
const T*
Signal<T>::
getSamplePtr (indexType x, indexType band) const
{
  return (T*) (this->sampleIndexX[x] + this->sampleIndexBand[band]);
}

template<class T>
inline
const T*
Signal<T>::
getSamplePtr (indexType x) const
{
  return (T*) this->sampleIndexX[x];
}


template<class T>
inline
T*
Signal<T>::
getTilePtr (indexType x, indexType band)
{
  return (T*) (this->tileIndexX[x] + this->tileIndexBand[band]);
}

template<class T>
inline
T*
Signal<T>::
getTilePtr (indexType x)
{
  return (T*) this->tileIndexX[x];
}

/*!

  Returns a list of neighbours

  \param pos position for which neighbours should be determined
  \param ng not relevant in this case...
 */
template<class T>
list<SignalPosition>
Signal<T>::getNeighbours(const SignalPosition &pos, const int ng) const throw (MocaException)
{
    indexType x = pos[DIM_X];
    list<SignalPosition> temp;

    if (ng != 2) {
        throw MocaException("Neighbour relation unknown.");
    } // if

    if (x > 0) {
        temp.push_front(SignalPosition(x - 1));
    } // if

    if (x < getSize()[DIM_X] - 1) {
        temp.push_front(SignalPosition(x + 1));
    } // if
} // Signal<T>::getNeighbours


/*
// find the position of the minimum value in the sampleblock
template<class T>
inline void Signal<T>::minValue (SignalPosition &pos)
{
	SampleBlockPosition t;
	SampleBlock<T>::minValue(t);
	pos[DIM_X]   = t[DIM_X];
	pos[DIM_Y]   = 0;
	pos[DIM_Z]   = 0;
	pos[DIM_BAND]= t[DIM_BAND];
	return;
}


// find the position of the maximum value in the sampleblock
template<class T>
inline void Signal<T>::maxValue (SignalPosition &pos)
{
	SampleBlockPosition t;
	SampleBlock<T>::maxValue(t);
	pos[DIM_X]   = t[DIM_X];
	pos[DIM_Y]   = 0;
	pos[DIM_Z]   = 0;
	pos[DIM_BAND]= t[DIM_BAND];
	return;
}
*/

// handles a vector of signals
template<class T>
class SignalVector
{
public: 
	SignalVector (int32 n) {
		size = n;
		data.resize(size);
	} 
private:
	int32 size;  // number of signals
public:
	vector<Signal<T> > data;
};

#endif
